"""Straw module

Straw enables programmatic access to .hic files.
.hic files store the contact matrices from Hi-C experiments and the
normalization and expected vectors, along with meta-data in the header.

The main function, straw, takes in the normalization, the filename or URL,
chromosome1 (and optional range), chromosome2 (and optional range),
whether the bins desired are fragment or base pair delimited, and bin size.

It then reads the header, follows the various pointers to the desired matrix
and normalization vector, and stores as [x, y, count]

Usage: straw <NONE/VC/VC_SQRT/KR> <hicFile(s)> <chr1>[:x1:x2] <chr2>[:y1:y2] <\
BP/FRAG> <binsize>

Example:
>>>import straw
>>>result = straw.straw('NONE', 'HIC001.hic', 'X', 'X', 'BP', 1000000)
>>>for i in range(len(result[0])):
...   print("{0}\t{1}\t{2}".format(result[0][i], result[1][i], result[2][i]))

See https://github.com/theaidenlab/straw/wiki/Python for more documentation
"""


__author__ = "Yue Wu and Neva Durand"
__license__ = "MIT"

import sys
import struct
import zlib
import requests
try:
    from io import StringIO
except ImportError:
    from io import StringIO

blockMap = dict()
# global version
version=0

def __readcstr(f):
    """ Helper function for reading in C-style string from file
    """
    buf = b""
    while True:
        b = f.read(1)
        if b is None or b == b"\0":
            return buf.decode("utf-8")
        elif b == "":
            raise EOFError("Buffer unexpectedly empty while trying to read null-terminated string")
        else:
            buf += b

def readHeader(req, chr1, chr2, posilist):
    """ Reads the header

    Args:
       req (file): File to read from
       chr1 (str): Chromosome 1
       chr2 (str): Chromosome 2
       c1pos1 (int, optional): Starting range of chromosome1 output
       c1pos2 (int, optional): Stopping range of chromosome1 output
       c2pos1 (int, optional): Starting range of chromosome2 output
       c2pos2 (int, optional): Stopping range of chromosome2 output

    Returns:
       list: master index, chromosome1 index, chromosome2 index
    """
    magic_string = struct.unpack('<3s', req.read(3))[0]
    req.read(1)
    if (magic_string != b"HIC"):
        print('This does not appear to be a HiC file magic string is incorrect')
        return -1
    global version
    version = struct.unpack('<i',req.read(4))[0]
    if (version < 6):
        print("Version {0} no longer supported".format(str(version)))
        return -1
    print('HiC version:' + '  {0}'.format(str(version)))
    master = struct.unpack('<q',req.read(8))[0]
    genome = b""
    c=req.read(1)
    while (c != b'\0'):
        genome += c
        c=req.read(1)

    # read and throw away attribute dictionary (stats+graphs)
    nattributes = struct.unpack('<i',req.read(4))[0]
    for x in range(nattributes):
        key = __readcstr(req)
        value = __readcstr(req)
    nChrs = struct.unpack('<i',req.read(4))[0]
    found1 = False
    found2 = False
    for i in range(0, nChrs):
      name = __readcstr(req)
      length = struct.unpack('<i',req.read(4))[0]
      if (name==chr1):
          found1=True
          chr1ind=i
          if (posilist[0]==-100):
              posilist[0]=0
              posilist[1]=length
      if (name==chr2):
          found2=True
          chr2ind=i
          if (posilist[2]==-100):
              posilist[2]=0
              posilist[3]=length
    if ((not found1) or (not found2)):
      print("One of the chromosomes wasn't found in the file. Check that the chromosome name matches the genome.\n")
      return -1
    return [master, chr1ind, chr2ind, posilist[0], posilist[1], posilist[2], posilist[3]]

def readFooter(req, c1, c2, norm, unit, resolution):
    """Reads the footer, which contains all the expected and normalization
    vectors. Presumes file pointer is in correct position
    Args:
       req (file): File to read from; presumes file pointer is in correct
       position
       chr1 (str): Chromosome 1
       chr2 (str): Chromosome 2
       norm (str): Normalization type, one of NONE, VC, KR, VC_SQRT
       unit (str): One of BP or FRAG
       resolution (int): Bin size

    Returns:
       list: File position of matrix, position+size chr1 normalization vector,
             position+size chr2 normalization vector
    """
    c1NormEntry=dict()
    c2NormEntry=dict()
    nBytes = struct.unpack('<i', req.read(4))[0]
    key = str(c1) + "_" + str(c2)
    nEntries = struct.unpack('<i', req.read(4))[0]
    found = False
    for i in range(nEntries):
        stri = __readcstr(req)
        fpos = struct.unpack('<q', req.read(8))[0]
        sizeinbytes = struct.unpack('<i', req.read(4))[0]
        if (stri==key):
            myFilePos = fpos
            found=True
    if (not found):
        print("File doesn't have the given chr_chr map\n")
    if (norm=="NONE"):
        return [myFilePos, 0, 0]
    nExpectedValues = struct.unpack('<i',req.read(4))[0]
    for i in range(nExpectedValues):
        str_ = __readcstr(req)
        binSize = struct.unpack('<i',req.read(4))[0]
        nValues = struct.unpack('<i',req.read(4))[0]
        for j in range(nValues):
            v = struct.unpack('<d',req.read(8))[0]
        nNormalizationFactors = struct.unpack('<i',req.read(4))[0]
        for j in range(nNormalizationFactors):
            chrIdx = struct.unpack('<i',req.read(4))[0]
            v = struct.unpack('<d',req.read(8))[0]
    nExpectedValues = struct.unpack('<i',req.read(4))[0]
    for i in range(nExpectedValues):
        str_ = __readcstr(req)
        str_ = __readcstr(req)
        binSize = struct.unpack('<i',req.read(4))[0]
        nValues = struct.unpack('<i',req.read(4))[0]
        for j in range(nValues):
            v = struct.unpack('<d',req.read(8))[0]
        nNormalizationFactors = struct.unpack('<i',req.read(4))[0]
        for j in range(nNormalizationFactors):
            chrIdx = struct.unpack('<i',req.read(4))[0]
            v = struct.unpack('<d',req.read(8))[0]
    nEntries = struct.unpack('<i',req.read(4))[0]
    found1=False
    found2=False
    for i in range(nEntries):
        normtype = __readcstr(req)
        chrIdx = struct.unpack('<i',req.read(4))[0]
        unit1 = __readcstr(req)
        resolution1 = struct.unpack('<i',req.read(4))[0]
        filePosition = struct.unpack('<q',req.read(8))[0]
        sizeInBytes = struct.unpack('<i',req.read(4))[0]
        if (chrIdx==c1 and normtype==norm and unit1==unit and resolution1==resolution):
            c1NormEntry['position']=filePosition
            c1NormEntry['size']=sizeInBytes
            found1=True
        if (chrIdx==c2 and normtype==norm and unit1==unit and resolution1==resolution):
            c2NormEntry['position']=filePosition
            c2NormEntry['size']=sizeInBytes
            found2=True
    if ((not found1) or (not found2)):
        print("File did not contain {0} normalization vectors for one or both chromosomes at {1} {2}\n".format(norm, resolution, unit))
        return -1
    return [myFilePos, c1NormEntry, c2NormEntry]


def readMatrixZoomData(req, myunit, mybinsize):
    """ Reads the Matrix Zoom Data, which gives pointer list for blocks for
    the data. Presumes file pointer is in correct position

    Args:
       req (file): File to read from; presumes file pointer is in correct
       position
       myunit (str): Unit (BP or FRAG) we're searching for
       mybinsize (int): Resolution we're searching for

    Returns:
       list containing boolean indicating if we found appropriate matrix,
       and if so, the counts for the bins and columns
    """
    unit = __readcstr(req)
    temp = struct.unpack('<i',req.read(4))[0]
    temp2 = struct.unpack('<f',req.read(4))[0]
    temp2 = struct.unpack('<f',req.read(4))[0]
    temp2 = struct.unpack('<f',req.read(4))[0]
    temp2 = struct.unpack('<f',req.read(4))[0]
    binSize = struct.unpack('<i',req.read(4))[0]
    blockBinCount = struct.unpack('<i',req.read(4))[0]
    blockColumnCount = struct.unpack('<i',req.read(4))[0]
    storeBlockData = False
    #for the initial
    myBlockBinCount = -1
    myBlockColumnCount = -1
    if (myunit==unit and mybinsize==binSize):
        myBlockBinCount=blockBinCount
        myBlockColumnCount=blockColumnCount
        storeBlockData=True
    nBlocks = struct.unpack('<i',req.read(4))[0]
    for b in range(nBlocks):
        blockNumber = struct.unpack('<i',req.read(4))[0]
        filePosition = struct.unpack('<q',req.read(8))[0]
        blockSizeInBytes = struct.unpack('<i',req.read(4))[0]
        entry=dict()
        entry['size'] = blockSizeInBytes
        entry['position'] = filePosition
        if (storeBlockData):
            blockMap[blockNumber] = entry
    return [storeBlockData, myBlockBinCount, myBlockColumnCount]

def readMatrix(req, unit, binsize):
    """ Reads the matrix - that is, finds the appropriate pointers to block
    data and stores them. Needs to read through headers of zoom data to find
    appropriate matrix. Presumes file pointer is in correct position.

    Args:
       req (file): File to read from; presumes file pointer is in correct
       position
       unit (str): Unit to search for (BP or FRAG)
       binsize (int): Resolution to search for

    Returns:
       list containing block bin count and block column count of matrix
    """
    c1 = struct.unpack('<i',req.read(4))[0]
    c2 = struct.unpack('<i',req.read(4))[0]
    nRes = struct.unpack('<i',req.read(4))[0]
    i = 0
    found = False
    blockBinCount = -1
    blockColumnCount = -1
    while (i<nRes and (not found)):
        list1 = readMatrixZoomData(req, unit, binsize)
        found = list1[0]
        if(list1[1]!=-1 and list1[2]!=-1):
            blockBinCount = list1[1]
            blockColumnCount = list1[2]
        i=i+1
    if (not found):
        print("Error finding block data\n")
        return -1
    return [blockBinCount, blockColumnCount]

def getBlockNumbersForRegionFromBinPosition(regionIndices, blockBinCount, blockColumnCount, intra):
    """ Gets the block numbers we will need for a specific region; used when
    the range to extract is sent in as a parameter

    Args:
       regionIndices (array): Array of ints giving range
       blockBinCount (int): The block bin count of the matrix
       blockColumnCount (int): The block column count of the matrix
       intra: Flag indicating if this is an intrachromosomal matrix

    Returns:
       blockSet (set): A set of blocks to print
    """
    col1=int(regionIndices[0]/blockBinCount)
    col2=int((regionIndices[1]+1)/blockBinCount)
    row1=int(regionIndices[2]/blockBinCount)
    row2=int((regionIndices[3]+1)/blockBinCount)
    blocksSet=set()
    # print(str(col1)+"\t"+str(col2)+"\t"+str(row1)+"\t"+str(row2))
    for r in range(row1, row2+1):
        for c in range(col1, col2+1):
            blockNumber=r*blockColumnCount+c
            blocksSet.add(blockNumber)
    if (intra):
        for r in range(col1, col2+1):
            for c in range(row1, row2+1):
                blockNumber=r*blockColumnCount+c
                blocksSet.add(blockNumber)
    # print(str(blocksSet))
    return blocksSet

def readBlock(req, size):
    """ Reads the block - reads the compressed bytes, decompresses, and stores
    results in array. Presumes file pointer is in correct position.

    Args:
       req (file): File to read from. Presumes file pointer is in correct
       position
       size (int): How many bytes to read

    Returns:
       array containing row, column, count data for this block
    """
    compressedBytes = req.read(size)
    uncompressedBytes = zlib.decompress(compressedBytes)
    nRecords = struct.unpack('<i',uncompressedBytes[0:4])[0]
    v=[]
    global version
    if (version < 7):
        for i in range(nRecords):
            binX = struct.unpack('<i',uncompressedBytes[(12*i+4):(12*i+8)])[0]
            binY = struct.unpack('<i',uncompressedBytes[(12*i+8):(12*i+12)])[0]
            counts = struct.unpack('<f',uncompressedBytes[(12*i+12):(12*i+16)])[0]
            record = dict()
            record['binX'] = binX
            record['binY'] = binY
            record['counts'] = counts
            v.append(record)
    else:
        binXOffset = struct.unpack('<i',uncompressedBytes[4:8])[0]
        binYOffset = struct.unpack('<i',uncompressedBytes[8:12])[0]
        useShort = struct.unpack('<b',uncompressedBytes[12:13])[0]
        type_ = struct.unpack('<b',uncompressedBytes[13:14])[0]
        index=0
        if (type_==1):
            rowCount = struct.unpack('<h',uncompressedBytes[14:16])[0]
            temp = 16
            for i in range(rowCount):
                y = struct.unpack('<h',uncompressedBytes[temp:(temp+2)])[0]
                temp=temp+2
                binY = y + binYOffset
                colCount = struct.unpack('<h',uncompressedBytes[temp:(temp+2)])[0]
                temp=temp+2
                for j in range(colCount):
                    x = struct.unpack('<h',uncompressedBytes[temp:(temp+2)])[0]
                    temp=temp+2
                    binX = binXOffset + x
                    if (useShort==0):
                        c = struct.unpack('<h',uncompressedBytes[temp:(temp+2)])[0]
                        temp=temp+2
                        counts = c
                    else:
                        counts = struct.unpack('<f',uncompressedBytes[temp:(temp+4)])[0]
                        temp=temp+4
                    record = dict()
                    record['binX'] = binX
                    record['binY'] = binY
                    record['counts'] = counts
                    v.append(record)
                    index = index + 1
        elif (type_== 2):
            temp=14
            nPts = struct.unpack('<i',uncompressedBytes[temp:(temp+4)])[0]
            temp=temp+4
            w = struct.unpack('<h',uncompressedBytes[temp:(temp+2)])[0]
            temp=temp+2
            for i in range(nPts):
                row=int(i/w)
                col=i-row*w
                bin1=int(binXOffset+col)
                bin2=int(binYOffset+row)
                if (useShort==0):
                    c = struct.unpack('<h',uncompressedBytes[temp:(temp+2)])[0]
                    temp=temp+2
                    if (c != -32768):
                        record = dict()
                        record['binX'] = bin1
                        record['binY'] = bin2
                        record['counts'] = c
                        v.append(record)
                        index = index + 1
                else:
                    counts = struct.unpack('<f',uncompressedBytes[temp:(temp+4)])[0]
                    temp=temp+4
                    if (countsnot != 0x7fc00000):
                        record = dict()
                        record['binX'] = bin1
                        record['binY'] = bin2
                        record['counts'] = counts
                        v.append(record)
                        index = index + 1
    return v

def readNormalizationVector(req):
    """ Reads the normalization vector from the file; presumes file pointer is
    in correct position

    Args:
       req (file): File to read from; presumes file pointer is in correct
       position

    Returns:
      Array of normalization values

    """
    value = []
    nValues = struct.unpack('<i',req.read(4))[0]
    for i in range(nValues):
        d = struct.unpack('<d',req.read(8))[0]
        value.append(d)
    return value

def straw(norm, infile, chr1loc, chr2loc, unit, binsize):
    """ This is the main workhorse method of the module. Reads a .hic file and
    extracts the given contact matrix. Stores in an array in sparse upper
    triangular format: row, column, (normalized) count

    Args:
       norm(str): Normalization type, one of VC, KR, VC_SQRT, or NONE
       infile(str): File name or URL of .hic file
       chr1loc(str): Chromosome name and (optionally) range, i.e. "1" or "1:10000:25000"
       chr2loc(str): Chromosome name and (optionally) range, i.e. "1" or "1:10000:25000"
       unit(str): One of BP or FRAG
       binsize(int): Resolution, i.e. 25000 for 25K
    """
    # clear the global variable blockMap so that it won't keep the data from previous calls
    for blockNum in list(blockMap.keys()):
        blockMap.pop(blockNum)

    magic_string = ""
    if (infile.startswith("http")):
        # try URL first. 100K should be sufficient for header
        headers={'range' : 'bytes=0-100000', 'x-amz-meta-requester' : 'straw'}
        s = requests.Session()
        r=s.get(infile, headers=headers)
        if (r.status_code >=400):
            print("Error accessing " + infile)
            print("HTTP status code " + str(r.status_code))
            return -1
        req=StringIO.StringIO(r.content)
        myrange=r.headers['content-range'].split('/')
        totalbytes=myrange[1]
    else:
        req=open(infile, 'rb')

    if (not (norm=="NONE" or norm=="VC" or norm=="VC_SQRT" or norm=="KR")):
        print("Norm specified incorrectly, must be one of <NONE/VC/VC_SQRT/KR>\nUsage: straw <NONE/VC/VC_SQRT/KR> <hicFile(s)> <chr1>[:x1:x2] <chr2>[:y1:y2] <BP/FRAG> <binsize>\n")
        return -1
    if (not (unit=="BP" or unit=="FRAG")):
        print("Unit specified incorrectly, must be one of <BP/FRAG>\nUsage: straw <NONE/VC/VC_SQRT/KR> <hicFile(s)> <chr1>[:x1:x2] <chr2>[:y1:y2] <BP/FRAG> <binsize>\n")
        return -1
    c1pos1=-100
    c1pos2=-100
    c2pos1=-100
    c2pos2=-100
    chr1_arra = chr1loc.split(":")
    chr2_arra = chr2loc.split(":")
    chr1=chr1_arra[0]
    chr2=chr2_arra[0]
    if(len(chr1_arra)==3):
        c1pos1=chr1_arra[1]
        c1pos2=chr1_arra[2]
    if(len(chr2_arra)==3):
        c2pos1=chr2_arra[1]
        c2pos2=chr2_arra[2]

    list1 = readHeader(req, chr1, chr2, [c1pos1, c1pos2, c2pos1, c2pos2])

    master=list1[0]
    chr1ind=list1[1]
    chr2ind=list1[2]
    c1pos1=int(list1[3])
    c1pos2=int(list1[4])
    c2pos1=int(list1[5])
    c2pos2=int(list1[6])
    c1=min(chr1ind,chr2ind)
    c2=max(chr1ind,chr2ind)
    origRegionIndices=[]
    regionIndices=[]
    if (chr1ind > chr2ind):
        origRegionIndices.append(c2pos1)
        origRegionIndices.append(c2pos2)
        origRegionIndices.append(c1pos1)
        origRegionIndices.append(c1pos2)
        regionIndices.append(int(c2pos1/binsize))
        regionIndices.append(int(c2pos2/binsize))
        regionIndices.append(int(c1pos1/binsize))
        regionIndices.append(int(c1pos2/binsize))
    else:
        origRegionIndices.append(c1pos1)
        origRegionIndices.append(c1pos2)
        origRegionIndices.append(c2pos1)
        origRegionIndices.append(c2pos2)
        regionIndices.append(int(c1pos1/binsize))
        regionIndices.append(int(c1pos2/binsize))
        regionIndices.append(int(c2pos1/binsize))
        regionIndices.append(int(c2pos2/binsize))

    # Get footer: from master to end of file
    if (infile.startswith("http")):
        headers={'range' : 'bytes={0}-{1}'.format(master, totalbytes) , 'x-amz-meta-requester' : 'straw'}
        #print("Requesting {} bytes".format(int(totalbytes)-master))
        r=s.get(infile, headers=headers);
        #print("Received {} bytes".format(r.headers['Content-Length']))
        req=StringIO.StringIO(r.content);
    else:
        req.seek(master)

    list1 = readFooter(req, c1, c2, norm, unit, binsize)
    myFilePos=list1[0]
    c1NormEntry=list1[1]
    c2NormEntry=list1[2]

    if (norm != "NONE"):
        if (infile.startswith("http")):
            endrange='bytes={0}-{1}'.format(c1NormEntry['position'],c1NormEntry['position']+c1NormEntry['size'])
            headers={'range' : endrange, 'x-amz-meta-requester' : 'straw'}
            r=s.get(infile, headers=headers);
            req=StringIO.StringIO(r.content);
            c1Norm = readNormalizationVector(req)

            endrange='bytes={0}-{1}'.format(c2NormEntry['position'],c2NormEntry['position']+c2NormEntry['size'])
            headers={'range' : endrange, 'x-amz-meta-requester' : 'straw'}
            r=s.get(infile, headers=headers)
            req=StringIO.StringIO(r.content)
            c2Norm = readNormalizationVector(req)
        else:
            req.seek(c1NormEntry['position'])
            c1Norm = readNormalizationVector(req)
            req.seek(c2NormEntry['position'])
            c2Norm = readNormalizationVector(req)

    if (infile.startswith("http")):
        headers={'range' : 'bytes={0}-'.format(myFilePos), 'x-amz-meta-requester' : 'straw'}
        r=s.get(infile, headers=headers, stream=True)
        list1 = readMatrix(r.raw, unit, binsize)
    else:
        req.seek(myFilePos)
        list1 = readMatrix(req, unit, binsize)

    blockBinCount=list1[0]
    blockColumnCount=list1[1]
    blockNumbers = getBlockNumbersForRegionFromBinPosition(regionIndices, blockBinCount, blockColumnCount, c1==c2)
    yActual=[]
    xActual=[]
    counts=[]

    for i_set in (blockNumbers):
        idx=dict()
        if(i_set in blockMap):
            idx=blockMap[i_set]
        else:
            idx['size']=0
            idx['position']=0
        if (idx['size']==0):
            records=[]
        else:
            if (infile.startswith("http")):
                endrange='bytes={0}-{1}'.format(idx['position'], idx['position']+idx['size'])
                headers={'range' : endrange, 'x-amz-meta-requester' : 'straw'}
                r=s.get(infile, headers=headers);
                req=StringIO.StringIO(r.content);
            else:
                req.seek(idx['position'])
            records=readBlock(req, idx['size'])

        for j in range(len(records)):
            rec=records[j]
            x=rec['binX']*binsize
            y=rec['binY']*binsize
            c=rec['counts']
            if (norm != "NONE"):
                a=c1Norm[rec['binX']]*c2Norm[rec['binY']]
                if (a!=0.0):
                    c=(c/(c1Norm[rec['binX']]*c2Norm[rec['binY']]))
                else:
                    c="inf"
            if ((x>=origRegionIndices[0] and x<=origRegionIndices[1] and y>=origRegionIndices[2] and y<=origRegionIndices[3]) or ((c1==c2) and y>=origRegionIndices[0] and y<=origRegionIndices[1] and x>= origRegionIndices[2] and x<=origRegionIndices[3])):
	            xActual.append(x)
	            yActual.append(y)
	            counts.append(c)
    return [xActual, yActual, counts]

def printme(norm, infile, chr1loc, chr2loc, unit, binsize,outfile):
    """ Reads a .hic file and extracts and prints the given contact matrix
    to a text file

    Args:
       norm(str): Normalization type, one of VC, KR, VC_SQRT, or NONE
       infile(str): File name or URL of .hic file
       chr1loc(str): Chromosome name and (optionally) range, i.e. "1" or "1:10000:25000"
       chr2loc(str): Chromosome name and (optionally) range, i.e. "1" or "1:10000:25000"
       unit(str): One of BP or FRAG
       binsize(int): Resolution, i.e. 25000 for 25K
       outfile(str): Name of text file to write to
    """
    f = open(outfile, 'w')
    result = straw(norm, infile, chr1loc, chr2loc, unit, binsize)
    for i in range(len(result[0])):
        f.write("{0}\t{1}\t{2}\n".format(result[0][i], result[1][i], result[2][i]))
        #print("{0}\t{1}\t{2}".format(result[0][i], result[1][i], result[2][i]))
    f.close()



import os,sys
from HiNT.corelib import *
from pkg_resources import resource_filename
#samtools path, pairixPath,bgzip,

def runBWA(opts,dataInfo):
    fq1,fq2 = opts.hicdata
    outbam = os.path.join(opts.outdir, opts.name + '.bam')
    command = '%s mem -SP5M -t %s %s %s %s | %s view -Shb - > %s'%(opts.bwapath,str(opts.threads),opts.bwaIndex,fq1,fq2,opts.samtoolspath,outbam)
    Info("Alignment with BWA-mem")
    print(command)
    run_cmd(command)
    dataInfo['fastq'] = opts.hicdata
    dataInfo['bam'] = outbam

    return dataInfo

def runpairsamtools(dataInfo,opts):
    inbam = dataInfo['bam']
    chromsizef = os.path.join(opts.referencedir, '%s.len'%opts.genome)
    outpairsam = os.path.join(opts.outdir,opts.name + '.pairsam.gz')
    command = '%s view -h %s | %s parse -c %s -o %s --assembly %s'%(opts.samtoolspath,inbam,opts.pairsampath,chromsizef,outpairsam,opts.genome)
    print(command)
    run_cmd(command)
    dataInfo['pairsam']=outpairsam

    return dataInfo

def sortpairsam(dataInfo,opts):
    chromsizef = os.path.join(opts.referencedir, '%s.len'%opts.genome)
    pairsamf = dataInfo['pairsam']
    sortedpairsam = pairsamf.rstrip('.pairsam.gz') + '.sorted.pairsam.gz'
    tmpdir = os.path.join(opts.outdir,'tmp')
    if not os.path.isdir(tmpdir):
        os.mkdir(tmpdir)
    command1 = "%s sort -o %s --memory 20G --compress-program gzip --tmpdir %s %s "%(opts.pairsampath,sortedpairsam,tmpdir,pairsamf)
    print(command1)
    run_cmd(command1)

    validpairsam = pairsamf.rstrip('.pairsam.gz') + '_valid.sorted.pairsam.gz'
    command2 = "%s select --chrom-subset %s -o %s '(pair_type == %s) or (pair_type == %s) or (pair_type == %s)' %s"%(opts.pairsampath,chromsizef,validpairsam, '"UU"','"UR"','"RU"',sortedpairsam)
    print(command2)
    run_cmd(command2)

    chimericpairsam = pairsamf.rstrip('.pairsam.gz') + '_chimeric.sorted.pairsam.gz'
    command3 = "%s select -o %s '(pair_type == %s) or (pair_type == %s) or (pair_type == %s)' %s"%(opts.pairsampath,chimericpairsam, '"NR"','"CC"','"MR"',sortedpairsam)
    print(command3)
    run_cmd(command3)

    validdeduppairsam = pairsamf.rstrip('.pairsam.gz') + '_valid.sorted.deduped.pairsam.gz'
    command4 = '%s dedup --output %s %s'%(opts.pairsampath,validdeduppairsam,validpairsam)
    print(command4)
    run_cmd(command4)

    validpairs = pairsamf.rstrip('.pairsam.gz') + '_merged_valid.pairs' #output pairs format
    command5 = '%s split --output-pairs %s %s'%(opts.pairsampath,validpairs,validdeduppairsam)
    print(command5)
    run_cmd(command5)

    os.remove(pairsamf)
    os.remove(validpairsam)
    os.remove(validdeduppairsam)

    dataInfo['validPairs'] = validpairs
    dataInfo['chimericPairsam'] = chimericpairsam

    return dataInfo

def pairsIndex(dataInfo,opts):
    validpairs = dataInfo["validPairs"]
    command1 = "bgzip -f %s"%(validpairs)
    print(command1)
    run_cmd(command1)

    command2 = "pairix -f %s"%(validpairs+'.gz')
    print(command2)
    run_cmd(command2)
    dataInfo['bgzippedValidPairs'] = validpairs+'.gz'

    return dataInfo

def runcooler(dataInfo,opts):
    pairsfile = dataInfo['bgzippedValidPairs']
    chromsizef = os.path.join(opts.referencedir, '%s.len'%opts.genome)
    if not opts.resolution:
        resolutions = [50,100,1000]
    else:
        resolutions = [50,100,1000] + [opts.resolution]

    for i in range(len(resolutions)):
        binsizefile = os.path.join(opts.outdir,"bins.%skb.bed"%str(resolutions[i]))
        command1 = '%s makebins %s %s > %s'%(opts.coolerpath,chromsizef,resolutions[i]*1000,binsizefile)
        print(command1)
        run_cmd(command1)
        outcool = os.path.join(opts.outdir,opts.name+'_%skb.cool'%str(resolutions[i]))
        command2 = '%s cload pairix %s %s %s'%(opts.coolerpath,binsizefile,pairsfile,outcool)
        print(command2)
        run_cmd(command2)
        command3 = 'cooler balance %s'%(outcool)
        print(command3)
        run_cmd(command3)
        dataInfo['%skb_cool'%(str(resolutions[i]))] = outcool

    return dataInfo

def runjuicer(dataInfo,opts):
    #Juicer Tools Version 1.7.6, java is required to be installed in advance
    pairsfile = dataInfo['bgzippedValidPairs']
    outhic = os.path.join(opts.outdir,opts.name+'.hic')
    command = 'java -jar %s pre -n %s %s %s'%(opts.juicerpath,pairsfile,outhic,opts.genome)
    print(command)
    run_cmd(command)
    command2 = 'java -jar %s addNorm -d -w 5000 -F %s'%(opts.juicerpath,outhic)
    print(command2)
    run_cmd(command2)
    dataInfo['hic'] = outhic

    return dataInfo



import os,sys
def readchemiricBPs(hicbpf):
	inf = open(hicbpf)
	Info = {}
	for line in inf:
		if line.startswith('chromosome1'):
			pass
		else:
			line = line.strip().split('\t')
			chr1,sr1,bp1,chr2,sr2,bp2,r1,r2,rt = line
			chrompair = chr1 + '_' + chr2
			if chrompair not in Info:
				Info[chrompair]=[[sr1,sr2,bp1,bp2,r1,r2,rt]]
			else:
				Info[chrompair].append([sr1,sr2,bp1,bp2,r1,r2,rt])
	# Info = {chrompair1:[[sr1,sr2,bp1,bp2,r1,r2,rt],[...],...],chrompair2:[[...],[...]]}
	return Info

def mergeCloseBPs(Info):
	#merge breakpoints that located within a 20bp windown centered from the compared one
	mergedInfo = {}
	for chrompair in Info:
		print(chrompair)
		bps = Info[chrompair]
		bps_sorted = sorted(bps, key=lambda x:(int(x[2]),int(x[3])))
		merged_bps = []
		merged_bps.append(bps_sorted[0])
		for i in range(1,len(bps_sorted)):
			sr11,sr21,bp11,bp21,r11,r21,rt1 = merged_bps[-1]
			sr12,sr22,bp12,bp22,r12,r22,rt2 = bps_sorted[i]
			if abs(int(bp12) - int(bp11) <= 10) and abs(int(bp22) - int(bp21) <= 10):
				if int(rt2) > int(rt1):
					merged_bps.remove(merged_bps[-1])
					merged_bps.append(bps_sorted[i])
				else:
					pass
			else:
				merged_bps.append(bps_sorted[i])
		sorted_mergedbps = sorted(merged_bps, key=lambda x:(int(x[-1])))
		mergedInfo[chrompair] = sorted_mergedbps

	return mergedInfo

def readchimericInfo(mergedInfo):
	chimericInfo = {}
	for k in mergedInfo:
		chr1,chr2 = k.split('_')
		values = mergedInfo[k]
		for value in values:
			searchregion1,searchregion2,bp1,bp2,supportread1,supportread2,supportreadtotal = value
			res = [chr1,searchregion1,bp1,chr2,searchregion2,bp2,supportread1,supportread2,supportreadtotal]
		try:
			chimericInfo[k].append([chr1,searchregion1,bp1,chr2,searchregion2,bp2,supportread1,supportread2,supportreadtotal])
		except:
			chimericInfo[k] = [[chr1,searchregion1,bp1,chr2,searchregion2,bp2,supportread1,supportread2,supportreadtotal]]

	return chimericInfo

def readGiniIndexf(giniIndexf):
	giniInfo = {}
	inf = open(giniIndexf)
	for line in inf:
		if line.startswith('GiniIndex'):
			pass
		else:
			line = line.strip().split('\t')
			chrompair, gini, maximum, rankproduct = line
			giniInfo[chrompair] = rankproduct
	return giniInfo

def readbpregionf(bpregionf):
	bpregionInfo = {}
	inf = open(bpregionf)
	for line in inf:
		line = line.strip().split('\t')
		chrompair,chr1,bp1,chr2,bp2 = line
		try:
			bpregionInfo[chrompair].append([chr1,bp1,chr2,bp2])
		except:
			bpregionInfo[chrompair] = [[chr1,bp1,chr2,bp2]]
	return bpregionInfo

def integration(bpregionInfo,chimericInfo,giniInfo,outbpf,resolution):
	outf = open(outbpf,'w')

	allres = []
	for chrompair in bpregionInfo:
		record = []
		regionbps = bpregionInfo[chrompair]
		rankproduct = giniInfo[chrompair]
		if chrompair not in chimericInfo:
			for regionbp in regionbps:
				chr1,bp1,chr2,bp2 = regionbp
				region1 = [str((int(bp1)-1)*resolution),str((int(bp1)+1)*resolution)]
				region2 = [str((int(bp2)-1)*resolution), str((int(bp2)+1)*resolution)]
				res = [chr1,'_'.join(region1),'-',chr2,'_'.join(region2),'-','0','0','0']
				allres.append((res + [str("%.2e" % float(rankproduct))]))
		else:
			chimericbps = chimericInfo[chrompair]
			for regionbp in regionbps:
				chr1,bp1,chr2,bp2 = regionbp
				regionbp1 = int(bp1)*resolution
				regionbp2 = int(bp2)*resolution
				n = 0
				for chimericbp in chimericbps:
					search_interval1 = chimericbp[1].split('_')
					search_interval2 = chimericbp[4].split('_')
					search_interval1 = [int(i) for i in search_interval1]
					search_interval2 = [int(i) for i in search_interval2]
					if (regionbp1 > search_interval1[0] and regionbp1 < search_interval1[1]) and (regionbp2 > search_interval2[0] and regionbp2 < search_interval2[1]):
						if chimericbp not in record:
							res = chimericbp
							allres.append((res + [str("%.2e" % float(rankproduct))]))
							n += 1
				if n ==0:
					region1 = [str((int(bp1)-1)*resolution),str((int(bp1)+1)*resolution)]
					region2 = [str((int(bp2)-1)*resolution), str((int(bp2)+1)*resolution)]
					res = [chr1,'_'.join(region1),'-',chr2,'_'.join(region2),'-','0','0','0']
					allres.append((res + [str("%.2e" % float(rankproduct))]))
	allres_sorted = sorted(allres, key=lambda x: (-float(x[-1]),float(x[-2])), reverse=True)

	title = ['chrom1','region1','breakpoint1','chrom2','region2','breakpoint2','supportRead1','supportRead2',"supportReadTotal","RankProduct(p-value)"]
	outf.write('\t'.join(title) + '\n')

	written = []
	for ar in allres_sorted:
		if ar not in written:
			outf.write('\t'.join(ar) + '\n')
			written.append(ar)
		else:
			pass
	outf.close()

def getSummarizedBP(validChimericbpf,outdir,name,giniIndexf,BPregionf):
	resolution = 100000
	summarizedChimericBPf = os.path.join(outdir,name + '_summarizedBP_fromChimeras.txt')
	chimericBPInfo = readchemiricBPs(validChimericbpf)
	mergedInfo = mergeCloseBPs(chimericBPInfo)
	FormatedChimericInfo = readchimericInfo(mergedInfo)
	giniInfo = readGiniIndexf(giniIndexf)
	bpregionInfo = readbpregionf(BPregionf)
	IntegratedBPf = os.path.join(outdir,name + '_Translocation_IntegratedBP.txt')
	integration(bpregionInfo,FormatedChimericInfo,giniInfo,IntegratedBPf,resolution)
	return IntegratedBPf



import numpy as np
from HiNT.straw import *
import os,sys

def get_chromInfo(chromlf):
	chroms = []
	infos = {}
	inf = open(chromlf)
	for line in inf:
		line = line.strip().split('\t')
		infos[line[0]] = int(line[1])
		chroms.append(line[0])
	return chroms,infos

def getSumPerChrom(i, j, hicfile, binsize, chroms, chromInfo, sumInfo):
    chrom1 = chroms[i] #the primary chromosome
    chrom2 = chroms[j] #the supplementary chromosome
    chr1 = chrom1.lstrip('chr')
    chr2 = chrom2.lstrip('chr')
    result = straw('NONE', hicfile, str(chr1), str(chr2), 'BP', binsize)
    #x is the coordinate for i, and y is the coordinate for j
    if i < j:
        xs = np.divide(result[0],binsize)
        ys = np.divide(result[1],binsize)
    else:
        xs = np.divide(result[1],binsize)
        ys = np.divide(result[0],binsize)
    values = np.array(result[2])
    chrom1length = chromInfo[chrom1]

    for n in sumInfo:
        idx = np.where(xs == n)
        nvalues = values[idx]
        nsum = np.sum(nvalues)
        sumInfo[n] += nsum
    return sumInfo

def writeGenomeRowSums(sumInfo,outputname,name,baseIdx):
    outf = open(outputname,'w')
    header=['',name]
    outf.write('\t'.join(header) + '\n')
    binsIdx = list(sumInfo.keys())
    binsIdx.sort()
    for idx in binsIdx:
        newidx = idx + baseIdx
        res = [str(newidx), str(sumInfo[idx])]
        outf.write('\t'.join(res) + '\n')
    outf.close()
    lastIdx = binsIdx[-1] + baseIdx
    return lastIdx

def getGenomeRowSums(resolution, hicfile, chromlf, outputdir,name):
    rowSumFilesInfo = {}
    chroms,chromInfo = get_chromInfo(chromlf)
    binsize = resolution * 1000
    baseIdx = 0
    for i in range(len(chroms)-2):
        sumInfo = {}
        chrom1length = chromInfo[chroms[i]]
        binnumber = int(chrom1length/binsize) + 1
        for n in range(binnumber):
            sumInfo[n] = 0
        for j in range(len(chroms)-2):
            sumInfo = getSumPerChrom(i, j, hicfile, binsize, chroms, chromInfo, sumInfo)
        outputname = os.path.join(outputdir,name + '_%s_%skb_GenomeRowSums.txt'%(chroms[i],str(resolution)))
        writeGenomeRowSums(sumInfo,outputname,name,baseIdx)
        rowSumFilesInfo[chroms[i]] = outputname
    return rowSumFilesInfo



import os,sys
from HiNT.corelib import *
import numpy as np

def write_Rscript(Rscript,datafile,outname,outpdfname,regressionLog):
	Rscript += 'library(mgcv)\n'
	Rscript += 'inf <- "%s"\n'%datafile
	Rscript += 'sink("%s",append=T)\n'%regressionLog
	Rscript += 'inf\n'
	Rscript += 'sink()\n'
	Rscript += 'data <- read.table(inf,sep="\t",header=T)\n'
	Rscript += 'data <- data.frame(data)\n'
	Rscript += 'lm1 <- gam(rowsum ~ s(gcper) + s(mapscore) + s(cutSitesNumber),data=data, family=poisson(link=log))\n'
	Rscript += 'sink("%s",append=T)\n'%regressionLog
	Rscript += 'summary(lm1)\n'
	Rscript += 'sink()\n'
	Rscript += 'rs <- residuals.gam(lm1,type="working")\n'
	Rscript += 'write.table(rs,file="%s",quote=F,row.names=F,col.names=F)\n'%outname
	Rscript += 'pdf("%s")\n'%outpdfname
	Rscript += 'par(mfrow=c(2,2))\n'
	Rscript += 'plot(lm1)\n'
	Rscript += 'dev.off()\n'
	return Rscript

def DoRegression(name,outdir,regressionData):

	Rfile = os.path.join(outdir,name+'_dataForRegression',name+'_GAMregression.R')
	regressionLog = os.path.join(outdir,name+'_dataForRegression',name+'_GAMregression.log')
	Rscript = ''
	filep = regressionData
	filen = regressionData.rstrip('.txt')
	outname = filen+'_residuals.r'
	outpdfname = os.path.join(outdir, name + '_GAMPoisson.pdf')
	Rscript = write_Rscript(Rscript,filep,outname,outpdfname,regressionLog)
	outf = open(Rfile,'w')
	outf.write(Rscript)
	outf.close()
	command = "Rscript %s"%Rfile
	print(command)
	run_cmd(command)
	residualf = outname
	return residualf

def sepResidualsByChrom(residualf,chroms,regressionFileInfo):
	residualFileInfo = {}
	allresiduals = np.loadtxt(residualf)
	for chromfile in regressionFileInfo:
		infile = regressionFileInfo[chromfile]
		infileDir = os.path.dirname(infile)
		outfilename = os.path.basename(infile).replace('.txt','_residuals.r')
		residualDir = os.path.join(infileDir,"residualsPerChrom")
		if not os.path.isdir(residualDir):
			os.mkdir(residualDir)
		outfile = os.path.join(residualDir,outfilename)
		residualFileInfo[chromfile] = [infile,outfile]
	start = 0
	for chrom in chroms:
		if chrom not in residualFileInfo:
			pass
		else:
			infname,outfname = residualFileInfo[chrom]
			#print infname,outfname
			inf = np.loadtxt(infname)
			nrow,ncol = np.shape(inf)
			end = start + nrow
			chromresidual = allresiduals[start:end]
			np.savetxt(outfname, chromresidual,fmt='%.10f')
			start += nrow

	return residualFileInfo

def calculateResiduals(name,outdir,regressionData,regressionFileInfo,chroms):
	AllchromResidualf = DoRegression(name,outdir,regressionData)
	residualChromFilesInfo = sepResidualsByChrom(AllchromResidualf,chroms,regressionFileInfo)
	return residualChromFilesInfo



import os,sys
import cooler
import numpy as np
import pandas as pds
import scipy.sparse as sps
from multiprocessing import Pool
import argparse

def getBins(coolfile):
	binsInfo = {}
	chroms = coolfile.chroms()["name"][:]
	for chrom in chroms:
		idxarray = np.where(coolfile.bins()["chrom"][:] == chrom)
		chromstart = idxarray[0][0]
		chromend = idxarray[0][-1]
		binsInfo[chrom] = [chromstart,chromend]
	#print binsInfo
	return chroms,binsInfo

def getSumPerChunk(coolfile,start,end):
	length = end - start
	rowSums = np.zeros((length,1))
	#print np.shape(rowSums)
	binNum = np.shape(coolfile.bins())[0]
	size = 100
	steps = int(binNum/size)
	for s in range(steps):
		start2 = s*size
		end2 = (s+1)*size
		matrix = coolfile.matrix(balance=False)[start:end, start2:end2]
		rowSum = np.sum(matrix, axis=1)
		rowSum = np.reshape(rowSum, (len(rowSum),1))
		rowSums += rowSum
	if steps*size < binNum:
		lastmatrix = coolfile.matrix(balance=False)[start:end, steps*size:]
		rowSum = np.sum(lastmatrix, axis=1)
		rowSum = np.reshape(rowSum, (len(rowSum),1))
		rowSums += rowSum
	else:
		pass

	return rowSums

def writeGenomeRowSums(coolfile,transRowSum,chrom,outputname,name):
	print("Writing rowsums of %s!"%chrom)
	bins = coolfile.bins()[:]
	allbins = bins[bins['chrom']== chrom].ix[:,0:3]
	genomeRowSum = np.reshape(transRowSum,(len(allbins),1))
	dfrowsums = pds.DataFrame(genomeRowSum,columns=[name],index=allbins.index)
	dfrowsums.to_csv(outputname,sep="\t")

def calRowsums(params):
	coolfile,binsInfo,targetchrom,chunksize,name,outputname = params
	s,e = binsInfo[targetchrom]
	length = e - s + 1
	RowSums = np.zeros((length,1))
	steps = int(length/chunksize)
	#print(steps)
	rowsums = np.array([])
	if steps > 0:
		for j in range(0,steps-1):
			#print j
			start = s + j*chunksize
			end = s + (j+1)*chunksize
			#print start,end
			rowSum = getSumPerChunk(coolfile,start,end)
			rowsums = np.append(rowsums,rowSum)
		start = s + chunksize*(steps-1)
		if start > e:
			pass
		else:
			end = e
			rowSum = getSumPerChunk(coolfile,start,end+1)
			rowsums = np.append(rowsums,rowSum)
		rowsums = np.reshape(rowsums,(len(rowsums),1))
	else:
		rowsums = getSumPerChunk(coolfile,s,e+1)
		rowsums = np.reshape(rowsums,(len(rowsums),1))
	genomeRowSum = RowSums + rowsums
	#print(np.shape(genomeRowSum))
	writeGenomeRowSums(coolfile,genomeRowSum,targetchrom,outputname,name)
	outInfo = targetchrom + '\t' + outputname +'\n'
	return outInfo

def getallChromsRowSums(coolpath,name,outputdir,resolution,threads):
	chunksize=int(1000/int(resolution))
	coolfile = cooler.Cooler(coolpath)
	chroms,binsInfo = getBins(coolfile)
	rowSumFilesInfo = {}

	allparamsInfo = []
	for i,targetchrom in enumerate(chroms):
		outputname = os.path.join(outputdir,name + '_%s_%skb_GenomeRowSums.txt'%(targetchrom,str(resolution)))
		allparamsInfo.append([coolfile,binsInfo,targetchrom,chunksize,name,outputname])
		rowSumFilesInfo[targetchrom] = outputname
	results = []
	p = Pool(threads)
	result = p.map_async(calRowsums, allparamsInfo, callback=results.append)
	p.close()
	p.join()
	#print results

	return rowSumFilesInfo



import os,sys
from pkg_resources import resource_filename
from multiprocessing import Pool
import numpy as np
import pypairix
from HiNT.corelib import *

def extrac_nonHiCchimeric(chimericPairsam,restrictionSites,restrictionEnzyme,outputPrefix):
	extractionScript = resource_filename('HiNT', 'externalScripts/extract_nonHiC.pl')
	command = "perl %s -psam %s -sites %s -re %s -outprefix %s"%(extractionScript, chimericPairsam, restrictionSites, restrictionEnzyme, outputPrefix)
	run_cmd(command)
	outputChimeras = outputPrefix+'_3dedup.txt'
	return outputChimeras

def makePairFormat(read):
	outres = False
	line = read.strip().split('\t')
	if len(line) == 16:
		readID,chrom1,pos1,strand1,score1,cigar1,sequence1,seq_qual1,chrom2,pos2,strand2,score2,cigar2,sequence2,flag,alternatives = line
	else:
		readID,chrom1,pos1,strand1,score1,cigar1,sequence1,chrom2,pos2,strand2,score2,cigar2,sequence2 = line
	if cigar1 != '*' and cigar2 != '-':
		start1 = pos1
		end1 = int(pos1) + 1
		start2 = pos2
		end2 = int(pos2) + 1
		res = [readID,chrom1,start1,str(end1),chrom2,start2,str(end2),strand1,strand2,cigar1,cigar2]
		outres = '\t'.join(res) + '\n'
		#outf.write('\t'.join(res) + '\n')
		if alternatives != '-':
			alternatives = alternatives.lstrip('XA:Z:').rstrip(';').split(';')
			for i in range(len(alternatives)):
				[chrom2,strand_pos2,cigar2,mismatch2] = alternatives[i].split(',')
				stand2 = strand_pos2[0]
				start2 = int(strand_pos2[1:])
				end2 = start2 + 1
				res = [readID,chrom1,start1,str(end1),chrom2,str(start2),str(end2),strand1,strand2,cigar1,cigar2]
				outres += '\t'.join(res) + '\n'
	return outres

def getchromsize(chromlengthf):
	chromSizeInfo = {}
	inf = open(chromlengthf)
	for line in inf:
		line = line.strip().split('\t')
		chrom,size = line
		chromSizeInfo[chrom] = int(size)
	return chromSizeInfo

def readFilteredBPs(breakpointsf,chromSizeInfo):
	bpInfo = {}
	inf = open(breakpointsf)
	for line in inf:
		if line.startswith('chrom'):
			pass
		else:
			line = line.strip().split('\t')
			chrompair,chrom1,bp1,chrom2,bp2 = line
			try:
				bpInfo[chrompair].append([(chrom1,bp1),(chrom2,bp2)])
			except:
				bpInfo[chrompair] = [[(chrom1,bp1),(chrom2,bp2)]]
	#print bpInfo
	return bpInfo

def is_qualified_clipped(cigar, cutoff_len):
	l=len(cigar)
	signal=[]
	lenth=[]
	temp=""
	for i in range(l):
		if cigar[i]>="0" and cigar[i]<="9":
			temp=temp+cigar[i]
		else:
			signal.append(cigar[i])
			try:
				lenth.append(int(temp))
			except ValueError:
				print("Error: ", cigar, temp)
			temp=""

	b_hardclip=False
	cnt_m=0
	clip_flag=0
	left_clip_len=0
	right_clip_len=0

	for i in range(len(signal)):
		if signal[i]=="M":
			cnt_m=cnt_m+lenth[i]

	if (signal[0]=="S" or signal[0]=="H") and lenth[0]>=cutoff_len:#left-clip
		clip_flag=1
		left_clip_len=lenth[0]
		if signal[0]=="H":
			b_hardclip=True
	if (signal[len(signal)-1]=="S" or signal[len(signal)-1]=="H") and lenth[len(signal)-1]>=cutoff_len: #right-clip
		clip_flag=clip_flag+2 #if this is 3, then both side clipped
		right_clip_len=lenth[len(signal)-1]
		if signal[len(signal)-1]=="H":
			b_hardclip=True

	return b_hardclip, cnt_m, clip_flag, left_clip_len, right_clip_len

def getclipPos_pair(read,m_clip_pos,reverse=False):
	CLIP_CUTOFF=13
	if reverse == True:
		ID,chromb,posb,endb,chroma,posa,enda,strandb,stranda,cigarb,cigara = read
	else:
		ID,chroma,posa,enda,chromb,posb,endb,stranda,strandb,cigara,cigarb = read
	b_hardclip_a, cnt_m_a, clip_flag_a, left_clip_len_a, right_clip_len_a = is_qualified_clipped(cigara, CLIP_CUTOFF)
	b_hardclip_b, cnt_m_b, clip_flag_b, left_clip_len_b, right_clip_len_b = is_qualified_clipped(cigarb, CLIP_CUTOFF)
	clip_pos_a=int(posa)
	if right_clip_len_a>0:#right clip
		clip_pos_a=int(posa)+cnt_m_a-1
	clip_pos_b = int(posb)
	if right_clip_len_b>0:#right clip
		clip_pos_b=int(posb)+cnt_m_b-1
	pair_pos = (clip_pos_a,clip_pos_b)
	if pair_pos not in m_clip_pos:
		m_clip_pos[pair_pos]=0
	m_clip_pos[pair_pos]=m_clip_pos[pair_pos]+1

	return m_clip_pos

def singleSide_clip_pos_calculation(m_clip_pos_forward,m_clip_pos_reverse,window_size=10):
	#clip_pos_a and clip_pos_b will be validated only when reads that chimeric in chr1~chr8 has such clip pos, and reads that chimeric in chr8~chr1 have such clip position too.
	clip_Pos_statistics = {}

	forward_keys = list(m_clip_pos_forward.keys())
	reverse_keys = list(m_clip_pos_reverse.keys())
	for i in range(len(forward_keys)):
		posa,posb = forward_keys[i]
		t = 0
		for j in range(-window_size,window_size+1):
			temp_posa = posa + j
			for l in range(-window_size,window_size+1):
				temp_posb = posb + l
				temp_pos = (temp_posa,temp_posb)
				if temp_pos in reverse_keys:
					t += 1
					if temp_pos not in clip_Pos_statistics:
						clip_Pos_statistics[temp_pos] = [m_clip_pos_forward[(posa,posb)],m_clip_pos_reverse[temp_pos],m_clip_pos_forward[(posa,posb)]+m_clip_pos_reverse[temp_pos]]
					else:
						clip_Pos_statistics[temp_pos] = [x + y for x,y in zip(clip_Pos_statistics[temp_pos], [m_clip_pos_forward[(posa,posb)],m_clip_pos_reverse[temp_pos],m_clip_pos_forward[(posa,posb)]+m_clip_pos_reverse[temp_pos]])]
		if t == 0:
			if (posa, posb) not in clip_Pos_statistics:
				clip_Pos_statistics[(posa,posb)] = [m_clip_pos_forward[(posa,posb)],0,m_clip_pos_forward[(posa,posb)]+0]
			else:
				clip_Pos_statistics[(posa,posb)] = [x + y for x,y in zip(clip_Pos_statistics[(posa,posb)], [m_clip_pos_forward[(posa,posb)],0,m_clip_pos_forward[(posa,posb)]+0])]

	for i in range(len(reverse_keys)):
		posa,posb = reverse_keys[i]
		m = 0
		for j in range(-window_size,window_size+1):
			temp_posa = posa + j
			for l in range(-window_size,window_size+1):
				temp_posb = posb + l
				temp_pos = (temp_posa,temp_posb)
				if temp_pos in forward_keys:
					m += 1
					pass
		if m == 0:
			if (posa,posb) not in clip_Pos_statistics:
				clip_Pos_statistics[(posa,posb)] = [0, m_clip_pos_reverse[(posa,posb)], m_clip_pos_reverse[(posa,posb)]+0]
			else:
				clip_Pos_statistics[(posa,posb)] = [x + y for x,y in zip(clip_Pos_statistics[(posa,posb)], [0,m_clip_pos_reverse[(posa,posb)],m_clip_pos_reverse[(posa,posb)]+0])]

	return clip_Pos_statistics

def further_filtering(clip_Pos_statistics,bp1,bp2):
	#The input clip_pos_dic can be either double validated clip positions or single side validated clip positions
	resolution = 100000
	filtered_clip_pos_statistics = {}
	print(clip_Pos_statistics)
	for cp in clip_Pos_statistics:
		if ((clip_Pos_statistics[cp][-2] >= 1) and (clip_Pos_statistics[cp][-3] >= 1)) or (clip_Pos_statistics[cp][-1] >= 2):
			filtered_clip_pos_statistics[cp] = clip_Pos_statistics[cp]
		elif (abs(cp[0] - int(bp1)*resolution) <= 2*resolution) and (abs(cp[1]-int(bp2)*resolution) <= 2*resolution):
			filtered_clip_pos_statistics[cp] = clip_Pos_statistics[cp]
		else:
			pass
	return filtered_clip_pos_statistics

def getReads(bpInfo,chromSizeInfo,chimericReadPairs,outdir,outname,windowsize=5,window_size=10):
	#windowsize is for the break point region detected from Hi-C, window_size is for the sliding window searching size
	tb = pypairix.open(chimericReadPairs)
	output = os.path.join(outdir,outname + '.supportedReads.txt')
	outf = open(output,'w')
	output3 = os.path.join(outdir,outname + '.singleValidated_pairs_BPs.txt')
	outf3 = open(output3,'w')
	resolution = 100000
	for chrompair in bpInfo:
		print(chrompair)
		chrompair_res1 = []
		chrompair_res2 = []
		chrompair_res3 = []
		bps = bpInfo[chrompair]
		supportedReads =[]
		for bp_detail in bps:
			chrom1Info,chrom2Info = bp_detail
			chr1,bp1 = chrom1Info
			chr2,bp2 = chrom2Info
			start = max(0,(int(bp1)-windowsize)*resolution)
			end = min((int(bp1)+windowsize)*resolution,chromSizeInfo[chrom1Info[0]])
			start2 = max(0,(int(bp2)-windowsize)*resolution)
			end2 = min((int(bp2)+windowsize)*resolution,chromSizeInfo[chrom2Info[0]])
			it1 = tb.query2D(chrom1Info[0],start,end,chrom2Info[0],start2,end2)
			m_clip_pos_forward = {}
			m_clip_pos_reverse = {}
			tempreads = []
			for read in it1:
				readinfo = read[1:]
				if readinfo not in tempreads:
					m_clip_pos_forward = getclipPos_pair(read,m_clip_pos_forward)
					tempreads.append(readinfo)
				else:
					pass
				if readinfo not in supportedReads:
					supportedReads.append(readinfo)
			it2 = tb.query2D(chrom2Info[0],start2,end2,chrom1Info[0],start,end)
			for read in it2:
				readinfo = read[1:]

				if readinfo not in tempreads:
					m_clip_pos_reverse = getclipPos_pair(read,m_clip_pos_reverse,reverse=True)
					tempreads.append(readinfo)
				else:
					pass
				if readinfo not in supportedReads:
					supportedReads.append(readinfo)
			clip_pos_statistics = singleSide_clip_pos_calculation(m_clip_pos_forward,m_clip_pos_reverse,window_size=10)
			filtered_clip_pos_statistics = further_filtering(clip_pos_statistics,bp1,bp2)
			reg_a = '_'.join([str(start),str(end)])
			reg_b = '_'.join([str(start2),str(end2)])
			for cp in filtered_clip_pos_statistics:
				#print chrompair_res
				pos_a,pos_b = cp
				hit_forward,hit_reverse,hit_all = filtered_clip_pos_statistics[cp]
				if hit_all == 0:
					pass
				else:
					if cp not in chrompair_res1:
						chrompair_res1.append(cp)
						res = [chrom1Info[0],reg_a,str(pos_a),chrom2Info[0],reg_b,str(pos_b)] + [str(hit_forward),str(hit_reverse),str(hit_all)]
						outf3.write('\t'.join(res)+"\n")
					else:
						pass
		for sr in supportedReads:
			outf.write('\t'.join(sr) + '\n')

	outf3.close()
	outf.close()
	return output3

def getBPfromChimeras(chimericPairsam,restrictionSites,restrictionEnzyme,outdir,name,chromlengthf,validbreakpointsf,pairixpath):
	chimericDir = os.path.join(outdir,'BPfromChimeric')
	if not os.path.isdir(chimericDir):
		os.mkdir(chimericDir)
	chimerasPrefix = os.path.join(chimericDir,name+'_nonHiC_chimeric')
	outputChimeras = extrac_nonHiCchimeric(chimericPairsam,restrictionSites,restrictionEnzyme,chimerasPrefix)
	inf = open(outputChimeras)
	Infos = []
	outputChimeraPairs = os.path.join(chimericDir,name+'_nonHiC_chimeric_3dedup.pairs')
	pool = Pool(16)
	with open(outputChimeraPairs,'w') as outf:
		for res in pool.imap(makePairFormat, inf):
			if res != False:
				outf.write(res)
	inf.close()
	sortedChimeras = os.path.join(chimericDir,name+'_nonHiC_chimeric_3dedup.bsorted.pairs.gz')
	command1 = "sort -k2,2 -k5,5 -k3,3n -k 6,6n %s | bgzip -c > %s"%(outputChimeraPairs,sortedChimeras)
	print(command1)
	run_cmd(command1)
	command2 = "%s -f -s2 -d5 -b3 -e4 -u6 -v7 %s"%(pairixpath,sortedChimeras)
	print(command2)
	run_cmd(command2)
	chromInfo = getchromsize(chromlengthf)
	bpInfo = readFilteredBPs(validbreakpointsf,chromInfo)
	BP_fromChimerasfile = getReads(bpInfo,chromInfo,sortedChimeras,outdir,name,windowsize=5,window_size=10)
	return BP_fromChimerasfile



import os,sys
from HiNT.corelib import *

def opt_validate_hintpre(optparser):
    """Validate arguments from a OptParser object.
    """

    options = optparser.parse_args()
    if options.inputformat == "fastq" and (not options.bwapath or not options.bwaIndex):
        optparser.print_help()
        Info('ERROR: HiNT need the path to bwa and bwa Index to do the alignmet.')
        sys.exit(1)

    if options.inputformat == "cooler" and not options.coolerpath:
        optparser.print_help()
        Info('ERROR: HiNT need the path to cooler to create HI-C contact matrix.')
        sys.exit(1)

    inputdata = options.hicdata
    if options.inputformat == "fastq":
        inputdatas = inputdata.split(',')
        if len(inputdatas) != 2:
            Info('ERROR: Both mates of Hi-C data should be input, a comma ',' should be used to seperate two mates. E.g. hic_1.fastq.gz,hic_2.fastq.gz')
            sys.exit(1)
        else:
            if not os.path.isfile(inputdatas[0]):
                Info('ERROR: %s is not a valid path'%inputdatas[0])
                sys.exit(1)
            if not os.path.isfile(inputdatas[1]):
                Info('ERROR: %s is not a valid path'%inputdatas[1])
                sys.exit(1)
        options.hicdata = inputdatas

    elif options.inputformat == "bam":
        if not os.path.isfile(inputdata):
            Info('ERROR: %s is not a valid path'%inputdata)
            sys.exit(1)
    else:
        Info('ERROR: choose a valid format for input data, fastq or bam')
        sys.exit(1)

    if not options.name:
        options.name = "NA"
        Info("HiNT will use 'NA' as the prefix for all the ouput files")

    if not options.outdir:
        options.outdir = os.getcwd()
        options.outdir = os.path.join(options.outdir,'HiNTpre_OUTPUT')
        if not os.path.isdir(options.outdir):
            os.mkdir(options.outdir)
    else:
        if not os.path.isdir(options.outdir):
            os.mkdir(options.outdir)


    Info("Argument List: ")
    Info("Hi-C data = " + ', '.join(options.hicdata))
    Info("Input format = " + options.inputformat)
    Info("Output format = " + options.outputformat)
    Info("Genome = %s" %options.genome)
    Info("Name = " + options.name)
    Info("Output directory = " + options.outdir)

    return options

def opt_validate_hintcnv(optparser):
    """
    Validate arguments from a OptParser object.
    """

    options = optparser.parse_args()
    if not os.path.isfile(options.matrixfile):
        optparser.print_help()
        Info('ERROR: path to the matrixfile is not valid.')
        sys.exit(1)

    if not os.path.isdir(options.bicseq):
        optparser.print_help()
        Info('ERROR: path to the BICseq package')
        sys.exit(1)

    if not options.name:
        options.name = "NA"
        Info("HiNT will use 'NA' as the prefix for all the ouput files")

    if not options.outdir:
        options.outdir = os.getcwd()
        options.outdir = os.path.join(options.outdir,'HiNTcnv_OUTPUT')
        if not os.path.isdir(options.outdir):
            os.mkdir(options.outdir)
    else:
        if not os.path.isdir(options.outdir):
            os.mkdir(options.outdir)


    Info("Argument List: ")
    Info("Hi-C contact matrix = " + options.matrixfile)
    Info("Hi-C contact matrix format = " + options.format)
    Info("resolution = " + str(options.resolution) + ' kb')
    Info("Genome = " + options.genome)
    Info("BICseq directory = " + options.bicseq)
    Info("Name = " + options.name)
    Info("Output directory = " + options.outdir)

    return options

def opt_validate_hinttransl(optparser):
    """
    Validate arguments from a OptParser object.
    """

    options = optparser.parse_args()
    if options.format == "juicer" and not os.path.isfile(options.matrixfile):
        optparser.print_help()
        Info('ERROR: path to the matrixfile is not valid.')
        sys.exit(1)
    if options.format == "cooler":
        options.matrixfile = options.matrixfile.split(',')
        if len(options.matrixfile) != 2 or not os.path.isfile(options.matrixfile[0]) or not os.path.isfile(options.matrixfile[1]):
            optparser.print_help()
            Info('ERROR: path to the matrixfile is not valid.')
            sys.exit(1)
    if (options.format == "sparse" or options.format == "dense") and not os.path.isdir(options.matrixfile):
        optparser.print_help()
        Info('ERROR: path to the matrixfile is not valid, should be a directory.')
        sys.exit(1)

    if options.chimeric and not os.path.isfile(options.chimeric):
        Info('ERROR: %s is not a valid path'%options.chimeric)
        sys.exit(1)

    if not options.name:
        options.name = "NA"
        Info("HiNT will use 'NA' as the prefix for all the ouput files")

    if not options.outdir:
        options.outdir = os.getcwd()
        options.outdir = os.path.join(options.outdir,'HiNTtransl_OUTPUT')
        if not os.path.isdir(options.outdir):
            os.mkdir(options.outdir)
    else:
        if not os.path.isdir(options.outdir):
            os.mkdir(options.outdir)


    Info("Argument List: ")
    if options.format == "cooler":
        Info("Hi-C contact matrix = " + ', '.join(options.matrixfile))
    if options.format == "juicer":
        Info("Hi-C contact matrix = " + options.matrixfile)
    Info("Hi-C contact matrix format = " + options.format)
    if options.chimeric:
        Info("Hi-C chimeric read pairs = " + options.chimeric)
    Info("Genome = %s" %options.genome)
    Info("Name = " + options.name)
    Info("Output directory = " + options.outdir)

    return options



import os,sys
import numpy as np
from scipy.stats import rankdata
from operator import itemgetter

def readBackgroundMatrix(backgroundMatrixDir):
    print(backgroundMatrixDir)
    backgroundMatrixInfo = {}
    matrices = os.listdir(backgroundMatrixDir)
    for matrix in matrices:
        nameInfo = matrix.split('_')
        chrompair = nameInfo[2] + '_' + nameInfo[3]
        matrixf = os.path.join(backgroundMatrixDir,matrix)
        backgroundMatrixInfo[chrompair] = matrixf

    return backgroundMatrixInfo

def gini(x):
    # (Warning: This is a concise implementation, but it is O(n**2)
    # in time and memory, where n = len(x).  *Don't* pass in huge
    # samples!)

    # Mean absolute difference
    mad = np.nanmean(np.abs(np.subtract.outer(x, x)))
    # Relative mean absolute difference
    rmad = mad/np.nanmean(x)
    # Gini coefficient
    g = 0.5 * rmad
    return g

def getGini(mat1,mat2):
    matrix1 = np.genfromtxt(mat1,delimiter="\t")
    matrix2 = np.genfromtxt(mat2,delimiter="\t")
    matrix1[np.isfinite(matrix1)==0] = 0
    matrix2[np.isfinite(matrix2)==0] = 0
    rowsum1 = np.sum(matrix1,axis=1)
    rowsum2 = np.sum(matrix2,axis=1)
    colsum1 = np.sum(matrix1,axis=0)
    colsum2 = np.sum(matrix2,axis=0)
    ridx1 = np.where(rowsum1==0)
    cidx1 = np.where(colsum1==0)
    ridx2 = np.where(rowsum2==0)
    cidx2 = np.where(colsum2==0)
    ridx = np.union1d(ridx1[0], ridx2[0])
    cidx = np.union1d(cidx1[0], cidx2[0])

    temp1 = np.delete(matrix1,ridx,0)
    temp2 = np.delete(matrix2,ridx,0)
    selectedData1 = np.delete(temp1,cidx,1)
    selectedData2 = np.delete(temp2,cidx,1)

    average1 = np.mean(selectedData1)
    average2 = np.mean(selectedData2)
    tm1 = np.divide(selectedData1,average1)
    tm2 = np.divide(selectedData2,average2)
    division = np.divide(tm1,tm2)
    giniIndex = gini(np.asarray(division).reshape(-1))
    maximum = np.nanmax(np.asarray(division).reshape(-1))

    return giniIndex,maximum

def getRankProduct(matrix1MbInfo,background1MbInfo,outdir,name):
    rpout = os.path.join(outdir,name + '_chrompairs_rankProduct.txt')
    outf = open(rpout,'w')
    ginis = []
    maximums = []
    chrompairs = []
    for chrompair in matrix1MbInfo:
        #print chrompair
        matrix1 = matrix1MbInfo[chrompair]
        matrix2 = background1MbInfo[chrompair]
        giniIndex,maximum = getGini(matrix1,matrix2)
        chrompairs.append(chrompair)
        ginis.append(giniIndex)
        maximums.append(maximum)
    rankgini = len(ginis) - rankdata(ginis)
    rankmaximum = len(maximums) - rankdata(maximums)
    #print rankgini,rankmaximum
    rps = (np.divide(rankgini,len(ginis)*1.0))*(np.divide(rankmaximum,len(maximums)*1.0))
    result = np.stack((chrompairs,ginis,maximums,rps),axis=-1)
    sortedResult = sorted(result, key=itemgetter(-1))
    outf.write('\t'.join(['ChromPair',"GiniIndex","Maximum","RankProduct"]) + '\n')
    for res in sortedResult:
        chrompair, gini, maximum, rp = res
        newres = [chrompair, str(gini), str(maximum), str(rp)]
        outf.write('\t'.join(newres) + '\n')
    outf.close()
   
    return rpout

def getRPinfo(rpout):
    rpInfo = {}
    inf = open(rpout)
    for line in inf:
        if line.startswith('Chrompair'):
            pass
        else:
            line = line.strip().split('\t')
            chrompair, gini, maximum, rp = line
            rpInfo[chrompair] = rp
    return rpInfo



import os,sys
import numpy as np
import cooler

def getBins(coolfile):
	binsInfo = {}
	chroms = coolfile.chroms()["name"][:]
	for chrom in chroms:
		idxarray = np.where(coolfile.bins()["chrom"][:] == chrom)
		chromstart = idxarray[0][0]
		chromend = idxarray[0][-1]
		binsInfo[chrom] = [chromstart,chromend]
		#print chrom, (chromend - chromstart + 1)
	return binsInfo

def dumpMatrix(resolution,coolfile,binsInfo,chrom1,chrom2,outdir,name):

	chrom1start,chrom1end = binsInfo[chrom1]
	chrom2start,chrom2end = binsInfo[chrom2]
	matrixName = '_'.join([name, str(resolution)+'kb',chrom1,chrom2,"InterMap_matrix.txt"])
	matrixfile = os.path.join(outdir,matrixName)
	matrix = coolfile.matrix(balance=True)[chrom1start:(chrom1end + 1), chrom2start:(chrom2end + 1)]
	np.savetxt(matrixfile,matrix,fmt='%.5f',delimiter='\t')

	return matrixfile

def coolToMatrix(matrixFile,resolution,outdir,name):
	MatrixInfo = {}
	coolfile = cooler.Cooler(matrixFile)
	binsInfo = getBins(coolfile)
	rankedChroms = ["chr1", "chr2", "chr3", "chr4", "chr5", "chr6", "chr7", "chr8", "chr9", "chr10", "chr11", "chr12", "chr13", "chr14", "chr15", "chr16", "chr17", "chr18", "chr19", "chr20", "chr21", "chr22", "chrX", "chrY"]
	outdir = os.path.join(outdir,"InterMap_matrix")
	if not os.path.isdir(outdir):
		os.mkdir(outdir)
	for i in range(0,len(rankedChroms)-2):
		for j in range((i+1), len(rankedChroms)-1):
			chrom1 = rankedChroms[i]
			chrom2 = rankedChroms[j]
			#print chrom1,chrom2
			matrixfile = dumpMatrix(resolution,coolfile,binsInfo,chrom1,chrom2,outdir,name)
			MatrixInfo[chrom1+'_'+chrom2] = matrixfile
	return MatrixInfo



import os,sys
import numpy as np
from scipy.sparse import coo_matrix
from HiNT.straw import *

def get_chromInfo(chromlf):
	chroms = []
	infos = {}
	inf = open(chromlf)
	for line in inf:
		line = line.strip().split('\t')
		infos[line[0]] = int(line[1])
		chroms.append(line[0])
	return chroms,infos

def dumpMatrix(chrom1, chrom2, resolution, hicfile, chromInfo, outputMatrixFile):
    chrom1length = chromInfo[chrom1]
    chrom2length = chromInfo[chrom2]
    binsize = resolution*1000
    binnumber1 = np.divide(chrom1length,binsize) + 1
    binnumber2 = np.divide(chrom2length,binsize) + 1
    chr1 = chrom1.lstrip('chr')
    chr2 = chrom2.lstrip('chr')
    result = straw('KR', hicfile, str(chr1), str(chr2), 'BP', binsize)
    row = np.divide(result[0],binsize)
    col = np.divide(result[1],binsize)
    data = result[2]
    #print max(row),max(col),binnumber1,binnumber2
    res = coo_matrix((data, (row, col)), shape=(binnumber1, binnumber2)).toarray()
    np.savetxt(outputMatrixFile, res, fmt='%.5f', delimiter='\t')

def hicToMatrix(hicfile, resolution, chromlf, outputdir,name):
    MatrixInfo = {}
    chroms,chromInfo = get_chromInfo(chromlf)
    for i in range(len(chroms)-3):
        chrom1length = chromInfo[chroms[i]]
        for j in range(i+1,len(chroms)-2):
            chrom1 = chroms[i]
            chrom2 = chroms[j]
            outmatrixdir = os.path.join(outputdir,"InterMap_matrix")
            if not os.path.isdir(outmatrixdir):
                os.mkdir(outmatrixdir)
            outputname = os.path.join(outmatrixdir,name + '_%skb_%s_%s_InterMap_matrix.txt'%(resolution,chrom1,chrom2))
            dumpMatrix(chrom1, chrom2, resolution, hicfile, chromInfo,outputname)
            MatrixInfo[chrom1+'_'+chrom2] = outputname
    return MatrixInfo



#!/home/sw229/miniconda2/bin/env python

#########################
####     Su Wang      ###
####    1-11-2018     ###
####   version 1.0    ###
#########################

"""Script Description:

HiNT - HiC for copy Number variations and Translocations detection

Part1:
Hi-C data preprocessing (fastq or bam format).
1, align to the genome via bwa-mem if the data format is fastq (skip if the data format is bam)
2, create valid read pairs (.pairs format) and chimeric read pairs (.pairsam format) with pairsamtools
3, create contact matrix with cooler or juicer tools in .cool or .hic format

Part2:
HiNT for CNV detection
1, extract 50kb (default) unnormalized contact matrices (.cool/.hic) or with input as a dense or sparse matrix format
2, calculate 1D coverage, GC content, mappability, and the number of restriction cutting sites in each 50kb bin (as default)
3, get residuals from GAM with Poisson linked function
4, segmentation with BIC-seq segmentation algorithm
5, visualization of CNVs

Part3:
HiNT for translocations detection
1, extract normalized Hi-C contact matrices in 100kb and 1Mb resolution with .cool or .hic format, or with the input as dense or sparse matrix fotmat
2, calculate the adjusted contact matrices with the in-house background contact matrices
3, calculate the rank product for each 1Mb resolution inter-chromosomal contact matrice to define the translocated chromosomal pairs
4, calculate the breakpoints based on contact matrices in 100kb resolution using a 1D coverage profile
5, Integrate with chimeric read pairs, refine the breakpoints to 1bp resolution


This code is free software; you can redistribute it and/or modify it.

@version: $1.0$
@author: Su Wang
@contact: wangsu0623@gmail.com
"""

import sys, os, time
import re
from pkg_resources import resource_filename
from HiNT.ArgsValidator import *
from HiNT.corelib import *

def prerun(args):
    opts = opt_validate_hintpre(args)
    from HiNT.HiCprocessingPip import runBWA,runpairsamtools,sortpairsam,pairsIndex,runcooler,runjuicer
    dataInfo = {}
    if opts.inputformat == "fastq":
        Info("Align to %s via BWA mem"%opts.genome)
        dataInfo = runBWA(opts,dataInfo)
    else:
        Info("Input data is in bam format, skip alignment step")
        dataInfo['bam'] = opts.hicdata
    dataInfo = runpairsamtools(dataInfo,opts)
    dataInfo = sortpairsam(dataInfo,opts)
    dataInfo = pairsIndex(dataInfo,opts)
    if opts.outputformat == 'cooler':
        dataInfo = runcooler(dataInfo,opts)
    elif opts.outputformat == 'juicer':
        dataInfo = runjuicer(dataInfo,opts)
    return

def cnvrun(args):
    opts = opt_validate_hintcnv(args)
    from HiNT.prepare_regression import get_chromInfo,getGCpercent,getmappability,getRestrictionSitesInfo,getFragmentsNumber,Regression,getnonzeros,mergeAllchroms,prepareData
    from HiNT.DoRegressionAllchroms import write_Rscript,DoRegression,sepResidualsByChrom,calculateResiduals
    from HiNT.doBICseq import makebinFiles,BICseqPrepare,run_BICseq
    chromlf = os.path.join(opts.referencedir,'%s.len'%opts.genome)
    if opts.format == 'cooler':
        from HiNT.getGenomeRowSumsFromCool import getBins,getSumPerChunk,writeGenomeRowSums,calRowsums,getallChromsRowSums
        rowSumFilesInfo = getallChromsRowSums(opts.matrixfile,opts.name,opts.outdir,opts.resolution,opts.threads) #Calculate rowsums
    if opts.format == 'juicer':
        from HiNT.getGenomeRowSumsFromHiC import get_chromInfo,getSumPerChrom,writeGenomeRowSums,getGenomeRowSums
        rowSumFilesInfo = getGenomeRowSums(opts.resolution, opts.matrixfile, chromlf, opts.outdir,opts.name)

    binsize = opts.resolution
    GCPercent_1kb = os.path.join(opts.referencedir, '%s_1kb_GCPercent.bed'%opts.genome)
    mappablity_track = os.path.join(opts.referencedir, '%s_mappability_50mer.bdg.gz'%opts.genome)
    mappablity_trackIndex = os.path.join(opts.referencedir, '%s_mappability_50mer.bdg.gz.tbi'%opts.genome)
    restrictionSites = os.path.join(opts.referencedir, '%s_%s_enzymeSites.txt'%(opts.genome,opts.enzyme))
    chroms,regressionFileAllchroms,regressionChromFilesInfo = prepareData(opts.name,opts.outdir,opts.referencedir,chromlf,rowSumFilesInfo,binsize,GCPercent_1kb,mappablity_track,restrictionSites) #Prepare the other data Information for regression
    #print chroms
    resiudalChromFilesInfo = calculateResiduals(opts.name,opts.outdir,regressionFileAllchroms,regressionChromFilesInfo,chroms) #form fuction DoregressionAllchroms, get residuals per chromosome
    #Segmentation step
    BICseqPath = os.path.join(opts.bicseq, 'BICseq2-seg.pl')
    CNVplotPath = resource_filename('HiNT', 'externalScripts/plot.l2r.ms_zoom.R')
    bicseqOut,outfig,outfig2 = run_BICseq(opts.name,opts.outdir,opts.resolution,resiudalChromFilesInfo,chroms,BICseqPath,CNVplotPath,opts.genome)
    rowsumsFiles = list(rowSumFilesInfo.values())
    for rowsumsFile in rowsumsFiles:
        os.remove(rowsumsFile)
    Info("Done! Find your CNV results from %s; %s; and %s ;)"%(bicseqOut,outfig,outfig2))
    return

def translrun(args):
    opts = opt_validate_hinttransl(args)
    chromlengthf = os.path.join(opts.referencedir, '%s.len'%opts.genome)
    Info("Prepare Matrices!")
    from HiNT.getRankProduct import readBackgroundMatrix,gini,getGini,getRankProduct,getRPinfo
    from HiNT.getRoughBreakpoints import getDivisionMatrix,runBPcaller,getAllRoughBreakpoints,getchromsize,readBPs,mergeValidBPs,mergeBPs,relativefold,filtering,getValidRoughBP
    if opts.format == 'cooler':
        from HiNT.coolToMatrix import getBins,dumpMatrix,coolToMatrix
        matrixfile1Mb,matrixfile100kb = opts.matrixfile
        matrix1MbInfo = coolToMatrix(matrixfile1Mb,1000,opts.outdir,opts.name)
        matrix100kbInfo = coolToMatrix(matrixfile100kb,100,opts.outdir,opts.name)
    if opts.format == 'juicer':
        from HiNT.hicToMatrix import get_chromInfo,dumpMatrix,hicToMatrix
        matrix1MbInfo = hicToMatrix(opts.matrixfile, 1000, chromlengthf, opts.outdir, opts.name)
        matrix100kbInfo = hicToMatrix(opts.matrixfile, 100, chromlengthf, opts.outdir, opts.name)
    backgroundMatrix1MbDir = os.path.join(opts.backgroundInterChromMatrixDir,opts.genome,'1Mb')
    backgroundMatrix100kbDir = os.path.join(opts.backgroundInterChromMatrixDir,opts.genome,'100kb')
    Info("Get matrix Info!")
    background1MbInfo = readBackgroundMatrix(backgroundMatrix1MbDir)
    background100kbInfo = readBackgroundMatrix(backgroundMatrix100kbDir)
    rpoutfile = getRankProduct(matrix1MbInfo,background1MbInfo,opts.outdir,opts.name)
    Info("Rank Product Done: find rank product for each chromosomal pair from %s. ;)"%rpoutfile)
    rpInfo = getRPinfo(rpoutfile)
    RscriptBPcallerPath = resource_filename('HiNT', 'externalScripts/getBreakPoints2steps.R')
    validBPregionOutf = getValidRoughBP(chromlengthf,matrix100kbInfo,background100kbInfo,rpInfo,opts.outdir,opts.name,opts.cutoff,RscriptBPcallerPath,opts.threads)
    if not opts.chimeric:
        Info("Done! Find your translocation breakpoints file from %s. ;)"%validBPregionOutf)
    else:
        from HiNT.getBPfromChimericReads import extrac_nonHiCchimeric,makePairFormat,getchromsize,readFilteredBPs,is_qualified_clipped,getclipPos_pair,singleSide_clip_pos_calculation,further_filtering,getReads,getBPfromChimeras
        from HiNT.BPsummarization import readchemiricBPs,mergeCloseBPs,readchimericInfo,readGiniIndexf,readbpregionf,integration,getSummarizedBP
        restrictionSites = os.path.join(opts.referencedir, '%s_%s_enzymeSites.txt'%(opts.genome,opts.enzyme)) #This may need to be modified according to Dhawal's script
        BP_fromChimerasfile = getBPfromChimeras(opts.chimeric,restrictionSites,opts.enzyme,opts.outdir,opts.name,chromlengthf,validBPregionOutf,opts.pairixpath)
        IntegratedBPf = getSummarizedBP(BP_fromChimerasfile,opts.outdir,opts.name,rpoutfile,validBPregionOutf)
        Info("Done! Find your translocation summarized breakpoints file from %s. ;)"%IntegratedBPf)
    return True



import os,sys
from HiNT.corelib import *
from multiprocessing import Pool
import numpy as np
import math

def getDivisionMatrix(mat1,mat2,fname):
	observedMatrix = np.genfromtxt(mat1,delimiter="\t")
	expectedMatrix = np.genfromtxt(mat2,delimiter="\t")
	expectedFinitevalues = expectedMatrix[np.isfinite(expectedMatrix)]
	observedFinitevalues = observedMatrix[np.isfinite(observedMatrix)]
	expected = expectedMatrix/np.mean(expectedFinitevalues)
	observed = observedMatrix/np.mean(observedFinitevalues)
	expected[np.where(expected==0)]=1
	with np.errstate(all='ignore'): divisionMatrix = observed/expected
	np.savetxt(fname,divisionMatrix,fmt="%.3f",delimiter='\t')
	return

def runBPcaller(params):
	#print params
	RscriptBPcaller,outputsubdir,matrixfile,tempbpoutputfile = params
	#print RscriptBPcaller,outputsubdir,matrixfile,tempbpoutputfile
	command = "Rscript %s %s %s %s"%(RscriptBPcaller,outputsubdir,matrixfile,tempbpoutputfile)
	print(command)
	run_cmd(command)
	chrompair_bps = open(tempbpoutputfile).readlines()
	os.remove(tempbpoutputfile)
	return chrompair_bps

def getAllRoughBreakpoints(matrix100kbInfo,background100kbInfo,rpInfo,outdir,name,cutoff,RscriptBPcaller,threads):
	DivisionMatrixInfo = {}
	outputsubdir = os.path.join(outdir,"AdjustedMatrixFiles100kb")
	if not os.path.isdir(outputsubdir):
		os.mkdir(outputsubdir)

	for chrompair in matrix100kbInfo:
		if chrompair not in background100kbInfo:
			pass
		else:
			if float(rpInfo[chrompair]) <= float(cutoff):
				mat1 = matrix100kbInfo[chrompair]
				mat2 = background100kbInfo[chrompair]
				fname = os.path.join(outputsubdir,name + '_' + chrompair + '_DivisionMatrix.txt')
				getDivisionMatrix(mat1,mat2,fname)
				DivisionMatrixInfo[chrompair] = fname
			else:
				pass
	#print DivisionMatrixInfo
	bpoutputfile = os.path.join(outdir,name + '_roughBP_100kb.txt')
	matrixfiles = os.listdir(outputsubdir)
	allparamsInfo = []
	for matrixfile in matrixfiles:
		matrixfile_prefix = matrixfile.replace('_DivisionMatrix.txt','')
		tempbpoutputfile = os.path.join(outdir,matrixfile_prefix + '_roughBP_100kb.txt')
		allparamsInfo.append([RscriptBPcaller,outputsubdir,matrixfile,tempbpoutputfile])
	#print allparamsInfo[0]
	results = []
	p = Pool(threads)
	result = p.map_async(runBPcaller, allparamsInfo, callback=results.append)
	p.close()
	p.join()

	#print results
	with open(bpoutputfile,'w') as outf:
		for i,line in enumerate(results[0]):
			if i == 0:
				outf.write(line[0])
				outf.write(line[1])
			else:
				outf.write(line[1])
	return bpoutputfile,DivisionMatrixInfo

##################################################################################################
##############    scripts below this line are doing the breakpoints filtering   ##################

def getchromsize(chromlengthf,resolution):
	chromSizeInfo = {}
	inf = open(chromlengthf)
	for line in inf:
		line = line.strip().split('\t')
		chrom,size = line
		chromSizeInfo[chrom] = [int(size),int(int(size)/int(resolution)+1)]
	return chromSizeInfo

def readBPs(breakpointsf,chromSizeInfo):
	bpInfo = {}
	inf = open(breakpointsf)
	for line in inf:
		if line.startswith('chrom'):
			pass
		else:
			line = line.strip().split('\t')
			chrom,bprow,maxrow,bpcol,maxcol = line
			chroms = chrom.split('_')
			chr1,chr2 = chroms
			size1 = chromSizeInfo[chr1][0]
			size2 = chromSizeInfo[chr2][0]
			if size1 > size2:
				bpInfo[chrom] = [(chr1,bpcol,maxcol),(chr2,bprow,maxrow)]
			else:
				bpInfo[chrom] = [(chr1,bprow,maxrow),(chr2,bpcol,maxcol)]
	return bpInfo

def mergeValidBPs(validBPs,matrixfile):
	matrix = np.loadtxt(matrixfile)
	rowsum = np.nansum(matrix,axis=1)
	colsum = np.nansum(matrix,axis=0)

	tempBPs = []
	for validbp in validBPs:
		if validbp not in tempBPs:
			tempBPs.append(validbp)
		else:
			pass
	bppairs = [(i[1],i[3]) for i in tempBPs]
	chrompair = list(set([(i[0],i[2]) for i in tempBPs]))
	if len(chrompair) != 1:
		print(chrompair)

	#print tempBPs,chrompair

	bppairs.sort()
	bp1s = [int(i[0]) for i in bppairs]
	bp2s = [int(i[1]) for i in bppairs]

	differences1 = ['NA']+[(bp1s[i+1]-bp1s[i]) for i in range(len(bp1s)-1)]
	differences2 = ['NA']+[(bp2s[i+1]-bp2s[i]) for i in range(len(bp2s)-1)]

	#print differences1,differences2
	tempMerged = [i for i in bppairs]
	#print tempMerged
	for j in range(1,len(differences1)):
		if differences1[j] == 0 and differences2[j] == 1:
			#print colsum[bp2s[j-1]],colsum[bp2s[j]]
			if colsum[bp2s[j-1]] > colsum[bp2s[j]]:
				tempMerged.remove(bppairs[j])
			else:
				tempMerged.remove(bppairs[j-1])
		elif differences1[j] == 1 and differences2[j] == 0:
			#print rowsum[bp2s[j-1]] > rowsum[bp2s[j]]
			if rowsum[bp2s[j-1]] > rowsum[bp2s[j]]:
				tempMerged.remove(bppairs[j])
			else:
				tempMerged.remove(bppairs[j-1])
		else:
			pass

	#print tempMerged

	mergedValidBPs = []
	for bp in tempMerged:
		allinfo = [chrompair[0][0],str(bp[0]),chrompair[0][1],str(bp[1])]
		mergedValidBPs.append(allinfo)

	return mergedValidBPs

def mergeBPs(bps,matrix,axis,windowsize=5):
	bps = [int(i)-1 for i in bps] #python index starts with 0
	bps.sort()
	mergedbps = [str(bps[0])]
	for i in range(1,len(bps)):
		a = int(bps[i])
		b = int(mergedbps[-1])
		#print a,b
		#a is always larger than b
		if (a - b) <= windowsize:
			if axis == "row":
				suma = np.nansum(matrix[a,:])
				sumar = np.nansum(matrix[a+1,:])
				sumb = np.nansum(matrix[b,:])
				sumbr = np.nansum(matrix[b+1,:])
			elif axis == 'column':
				suma = np.nansum(matrix[:,a])
				sumar = np.nansum(matrix[:,a+1])
				sumb = np.nansum(matrix[:,b])
				sumbr = np.nansum(matrix[:,b+1])
			else:
				print("give the correct axis, row or column")
			if suma > sumar and sumb < sumbr: # __|--|__
				fca = suma/sumar
				fcb = sumbr/sumb
				if suma > sumbr and fca > fcb:
					print("merge",np.nansum(matrix[a-1,:]), suma, sumar, np.nansum(matrix[b-1,:]),sumb, sumbr, fca, fcb)
					mergedbps.append(str(a))
					mergedbps.remove(str(b))
				elif suma < sumbr and fca > fcb:
					mergedbps.append(str(a))
				else:
					continue
		else:
			mergedbps.append(str(a))
	return mergedbps

def relativefold(a,b):
	if float(b) != 0:
		rfold = (float(a) - float(b))/float(b)
	else:
		rfold = (float(a) - float(b))/(float(b)+1)
	return rfold

def filtering(bps,matrixfile,chromSizeInfo,windowsize):
	matrix = np.loadtxt(matrixfile)
	rowsum = np.nansum(matrix,axis=1)
	#print np.shape(rowsum)
	colsum = np.nansum(matrix,axis=0)
	#print np.shape(colsum)
	chrom1Info,chrom2Info = bps
	chr1,bp1,max1 = chrom1Info
	chr2,bp2,max2 = chrom2Info
	bp1 = bp1.split(',')
	bp2 = bp2.split(',')

	bp1s = bp1
	bp2s = bp2
	bp1s = list(set(bp1s))
	bp2s = list(set(bp2s))

	#print bp1s,bp2s
	mergedbp1s = mergeBPs(bp1s,matrix,'row')
	#print mergedbp1s
	if str(int(max1)-1) not in mergedbp1s:
		mergedbp1s = mergedbp1s + [str(int(max1)-1)]

	mergedbp2s = mergeBPs(bp2s,matrix,'column')
	#print mergedbp2s

	if str(int(max2)-1) not in mergedbp2s:
		mergedbp2s = mergedbp2s + [str(int(max2)-1)]

	print(mergedbp1s,mergedbp2s)

	validBPs = []
	binsize1 = chromSizeInfo[chr1][1]
	binsize2 = chromSizeInfo[chr2][1]

	cutoff = np.nanpercentile(matrix,99)
	#print "cutoff:", cutoff

	for bp1 in mergedbp1s:
		for bp2 in mergedbp2s:
			sm1 = np.nanmedian(matrix[int(bp1)-5:int(bp1),int(bp2)+1:int(bp2)+6])
			sm2 = np.nanmedian(matrix[int(bp1)+1:int(bp1)+6,int(bp2)+1:int(bp2)+6])
			sm3 = np.nanmedian(matrix[int(bp1)-5:int(bp1),int(bp2)-5:int(bp2)])
			sm4 = np.nanmedian(matrix[int(bp1)+1:int(bp1)+6,int(bp2)-5:int(bp2)])
			#print sm1,sm2,sm3,sm4
			sms = np.sort(np.asarray([sm1,sm2,sm3,sm4]))
			translocationType = "unbalanced"
			if sms[2] > min(np.nanpercentile(matrix,80),2) and sms[1] < max(np.nanpercentile(matrix,60),1.5):
				translocationType = "balanced"
			#print translocationType

			if translocationType == "NULL":
				continue
			else:

				if translocationType == "balanced":
					if sm3 > sm4: #bp1 belongs to the left
					#if rowsum[int(bp1)] > rowsum[int(bp1)+1]:
						x1 = int(bp1)-windowsize+1
						x2 = int(bp1)+1
					else:
						x1 = int(bp1)-windowsize
						x2 = int(bp1)
				if translocationType == "unbalanced":
					if np.nanmean(rowsum[int(bp1)-5:int(bp1)]) > np.nanmean(rowsum[int(bp1)+1:int(bp1)+6]):
						x1 = int(bp1)-windowsize+1
						x2 = int(bp1)+1
					else:
						x1 = int(bp1)-windowsize
						x2 = int(bp1)

				if translocationType == "balanced":
					if sm3 > sm1:
						#if colsum[int(bp2)] > colsum[int(bp2)+1]:
						y1 = int(bp2)-windowsize+1
						y2 = int(bp2)+1
					else:
						y1 = int(bp2)-windowsize
						y2 = int(bp2)

				if translocationType == "unbalanced":
					if np.nanmean(colsum[int(bp2)-5:int(bp2)]) > np.nanmean(colsum[int(bp2)+1:int(bp2)+6]):
						y1 = int(bp2)-windowsize+1
						y2 = int(bp2)+1
					else:
						y1 = int(bp2)-windowsize
						y2 = int(bp2)

				if x1 < 0:
					x1 = 0
				if y1 < 0:
					y1 = 0
				#print x1,x2,y1,y2
				r1 = matrix[x1:x2,y1:y2]
				mr1 = np.nanmean(r1)

				if translocationType == "balanced":
					if sm1 > sm2:
						x1 = int(bp1)+1
						x2 = int(bp1)+1+windowsize
					else:
						x1 = int(bp1)
						x2 = int(bp1)+windowsize
				if translocationType == "unbalanced":
					if np.nanmean(rowsum[int(bp1)-5:int(bp1)]) > np.nanmean(rowsum[int(bp1)+1:int(bp1)+6]):
						x1 = int(bp1)+1
						x2 = int(bp1)+1+windowsize
					else:
						x1 = int(bp1)
						x2 = int(bp1)+windowsize
				if translocationType == "balanced":
					if sm4 > sm2:
						y1 = int(bp2)+1
						y2 = int(bp2)+1+windowsize
					else:
						y1 = int(bp2)
						y2 = int(bp2)+windowsize
				if translocationType == "unbalanced":
					if np.nanmean(colsum[int(bp2)-5:int(bp2)]) > np.nanmean(colsum[int(bp2)+1:int(bp2)+6]):
						y1 = int(bp2)+1
						y2 = int(bp2)+1+windowsize
					else:
						y1 = int(bp2)
						y2 = int(bp2)+windowsize
				if x2 >= binsize1:
					x2 = binsize1
				if y2 >= binsize2:
					y1 = binsize2
				#print x1,x2,y1,y2
				r2 = matrix[x1:x2,y1:y2]
				mr2 = np.nanmean(r2)

				if translocationType == "balanced":
					if sm1 > sm2:
						x1 = int(bp1)-windowsize+1
						x2 = int(bp1)+1
					else:
						x1 = int(bp1)-windowsize
						x2 = int(bp1)
				if translocationType == "unbalanced":
					if np.nanmean(rowsum[int(bp1)-5:int(bp1)]) > np.nanmean(rowsum[int(bp1)+1:int(bp1)+6]):
						x1 = int(bp1)-windowsize+1
						x2 = int(bp1)+1
					else:
						x1 = int(bp1)-windowsize
						x2 = int(bp1)
				if translocationType == "balanced":
					if sm3 > sm1:
						y1 = int(bp2)+1
						y2 = int(bp2)+1+windowsize
					else:
						y1 = int(bp2)
						y2 = int(bp2)+windowsize
				if translocationType == "unbalanced":
					if np.nanmean(colsum[int(bp2)-5:int(bp2)]) > np.nanmean(colsum[int(bp2)+1:int(bp2)+6]):
						y1 = int(bp2)+1
						y2 = int(bp2)+1+windowsize
					else:
						y1 = int(bp2)
						y2 = int(bp2)+windowsize
				if x1 < 0:
					x1 = 0
				if y2 > binsize2:
					y1 = binsize2
				#print x1,x2,y1,y2
				r3 = matrix[x1:x2,y1:y2]
				mr3 = np.nanpercentile(r3,80,interpolation="lower")
				mr3 = np.nanmean(r3)

				if translocationType == "balanced":
					if sm3 > sm4:
						x1 = int(bp1)+1
						x2 = int(bp1)+1+windowsize
					else:
						x1 = int(bp1)
						x2 = int(bp1)+windowsize
				if translocationType == "unbalanced":
					if np.nanmean(rowsum[int(bp1)-5:int(bp1)]) > np.nanmean(rowsum[int(bp1)+1:int(bp1)+6]):
						x1 = int(bp1)+1
						x2 = int(bp1)+1+windowsize
					else:
						x1 = int(bp1)
						x2 = int(bp1)+windowsize
				if translocationType == "balanced":
					if sm4 > sm2:
						y1 = int(bp2)-windowsize+1
						y2 = int(bp2)+1
					else:
						y1 = int(bp2)-windowsize
						y2 = int(bp2)
				if translocationType == "unbalanced":
					if np.nanmean(colsum[int(bp2)-5:int(bp2)]) > np.nanmean(colsum[int(bp2)+1:int(bp2)+6]):
						y1 = int(bp2)-windowsize+1
						y2 = int(bp2)+1
					else:
						y1 = int(bp2)-windowsize
						y2 = int(bp2)

				if x2 >= binsize1:
					x2 = binsize1
				if y1 < 0:
					y1 = 0
				#print x1,x2,y1,y2
				r4 = matrix[x1:x2,y1:y2]
				#print(np.shape(r4))
				mr4 = np.nanpercentile(r4,80,interpolation="lower")
				mr4 = np.nanmean(r4)
				###   mr1 | mr3
				###   ----|-----
				###   mr4 | mr2

				#print mr1,mr2,mr3,mr4,cutoff
				count = 0
				mrs = [mr1,mr2,mr3,mr4]
				for mr in mrs:
					if np.isnan(mr):
						pass
					else:
						if mr > cutoff:
							count += 1
				#print count

				if count == 0:
					pass
				elif count >= 3:
					pass
				elif count == 2: #balanced translocation, need a submatrix to calculate the rowsums and colsums
					mrs.sort()
					if (mr1 > cutoff and mr2 > cutoff) or (mr3> cutoff and mr4 > cutoff):
						rfold1 = relativefold(mrs[-2],mrs[-3])
						#print rfold1
						if rfold1 > 5:
							#print 'TRUE',bp1,bp2
							submatrix = matrix[:,0:int(bp2)]
							subrowsum = np.nansum(submatrix,1)
							submatrix = matrix[0:int(bp1),:]
							subcolsum = np.nansum(submatrix,0)

							if subrowsum[int(bp1)] > subrowsum[int(bp1)+1]:
								correctbp1 = bp1
							else:
								correctbp1 = str(int(bp1)+1)
							if subcolsum[int(bp2)] > subcolsum[int(bp2)+1]:
								correctbp2 = bp2
							else:
								correctbp2 = str(int(bp2)+1)

							validBPs.append([chr1,str(int(correctbp1)+1),chr2,str(int(correctbp2)+1)])
					else:
						pass
				else:
					#print 'count: ', count
					sortedmrs = sorted(mrs, key = lambda x : float('-inf') if math.isnan(x) else x)
					#print sortedmrs
					rfold1 = 1
					rfold2 = 1
					if not math.isnan(sortedmrs[-2]) and sortedmrs[-2] != 0:
						rfold1 = relativefold(sortedmrs[-1],sortedmrs[-2])
						rfold2 = 1
						#print 'rfold1: ', rfold1
					else:
						rfold2 = relativefold(sortedmrs[-1],cutoff)
						#print 'rfold2: ', rfold2
						rfold1 = 1

					if rfold1 > 5 or rfold2 > 2:
						#print 'TRUE',bp1,bp2
						if rowsum[int(bp1)] > rowsum[int(bp1)+1]:
							correctbp1 = bp1
						else:
							correctbp1 = str(int(bp1)+1)
						if colsum[int(bp2)] > colsum[int(bp2)+1]:
							correctbp2 = bp2
						else:
							correctbp2 = str(int(bp2)+1)
						validBPs.append([chr1,str(int(correctbp1)+1),chr2,str(int(correctbp2)+1)])

	return validBPs

def getValidRoughBP(chromlengthf,matrix100kbInfo,background100kbInfo,rpInfo,outdir,name,cutoff,RscriptBPcaller,threads):
	bpoutputfile,DivisionMatrixInfo = getAllRoughBreakpoints(matrix100kbInfo,background100kbInfo,rpInfo,outdir,name,cutoff,RscriptBPcaller,threads)
	chromSizeInfo = getchromsize(chromlengthf,resolution = 100000)
	bpInfo = readBPs(bpoutputfile,chromSizeInfo)
	windowsize = 10
	validbpoutputfile = os.path.join(outdir,name + '_roughBP_100kb_filtered.txt')
	outf = open(validbpoutputfile,'w')
	#print DivisionMatrixInfo.keys()
	for chrompair in bpInfo:
		if chrompair in DivisionMatrixInfo:
			#print chrompair
			bps = bpInfo[chrompair]
			matrixfile = DivisionMatrixInfo[chrompair]
			validBPs = filtering(bps,matrixfile,chromSizeInfo,windowsize)
			if len(validBPs) == 1:
				mergedValidBPs = validBPs
			else:
				mergedValidBPs = mergeValidBPs(validBPs,matrixfile)
			for validbp in mergedValidBPs:
				res = [chrompair] + validbp
				outf.write('\t'.join(res) + '\n')
		else:
			print("This chromosomal pair (%s) is not valid"%chrompair)
	outf.close()

	return validbpoutputfile




# print current time and information on screen
import subprocess
from subprocess import call as subpcall
import sys, os, time
import logging

error   = logging.critical		# function alias
warn    = logging.warning
def Info(infoStr):
    print("[%s] %s" %(time.strftime('%H:%M:%S'), infoStr))
    
def run_cmd(command):
    subpcall (command, shell = True)



#Reads the genome name from the hic header
#Can take in a .hic file or URL that points to .hic file
import sys
import struct
import requests
import io

def readcstr(f):
    buf = bytearray()
    while True:
        b = f.read(1)
        if b is None or b == '\0':
            return str(buf)
        else:
            buf.append(b)

if (len(sys.argv) != 2):
  sys.stderr.write('Usage: '+ sys.argv[0]+' <hic file or URL>\n')
  sys.exit(1)

infile = sys.argv[1]
magic_string = ""

if (infile.startswith("http")):
    # try URL first. 100K should be sufficient for header
    headers={'range' : 'bytes=0-100000', 'x-amz-meta-requester' : 'straw'}
    s = requests.Session()
    r=s.get(infile, headers=headers)
    if (r.status_code >=400):
        print(("Error accessing " + infile)) 
        print(("HTTP status code " + str(r.status_code)))
        sys.exit(1)
    req=io.StringIO(r.content)        
    myrange=r.headers['content-range'].split('/')
    totalbytes=myrange[1]
else:
    req=open(infile, 'rb')

magic_string = struct.unpack('<3s', req.read(3))[0]
req.read(1)
if (magic_string != b"HIC"):
    print('This does not appear to be a HiC file magic string is incorrect')
    sys.exit(1)
version = struct.unpack('<i',req.read(4))[0]
print('HiC version:')
print('  {0}'.format(str(version))) 
masterindex = struct.unpack('<q',req.read(8))[0]
genome = ""
c=req.read(1)
while (c != '\0'):
    genome += c
    c=req.read(1)
print('Genome ID:')
print('  {0}'.format(str(genome))) 
# read and throw away attribute dictionary (stats+graphs)
print('Attribute dictionary:')
nattributes = struct.unpack('<i',req.read(4))[0]
for x in range(0, nattributes):
  key = readcstr(req)
  value = readcstr(req)
#  print '   Key:{0}'.format(key)
#  print '   Value:{0}'.format(value)
nChrs = struct.unpack('<i',req.read(4))[0]
print("Chromosomes: ")
for x in range(0, nChrs):
  name = readcstr(req)
  length = struct.unpack('<i',req.read(4))[0]
  print('  {0}  {1}'.format(name, length))
nBpRes = struct.unpack('<i',req.read(4))[0]
print("Base pair-delimited resolutions: ")
for x in range(0, nBpRes):
  res = struct.unpack('<i',req.read(4))[0]
  print('   {0}'.format(res))
nFrag = struct.unpack('<i',req.read(4))[0]
print("Fragment-delimited resolutions: ")
for x in range(0, nFrag):
  res = struct.unpack('<i',req.read(4))[0]
  print('   {0}'.format(res))



#We would like to regress rowsums with mappablity, GC content and the number of fragment is this region
import os,sys
import numpy as np
from sklearn import linear_model
import pandas as pd
import tabix
from HiNT.corelib import *
from pkg_resources import resource_filename

def get_chromInfo(chromlf):
	chroms = []
	infos = {}
	inf = open(chromlf)
	for line in inf:
		line = line.strip().split('\t')
		infos[line[0]] = line[1]
		chroms.append(line[0])
	return chroms,infos

def getGCpercent(gcfile,chrom,resolution):
	GCdata = pd.read_csv(gcfile, header = 0,sep="\t",index_col=0)
	GCper = GCdata.loc[chrom:chrom,'5_pct_gc']
	end = GCdata.loc[chrom:chrom,'3_usercol']
	length = end[-1:][0]
	steps = int(length/resolution)
	GC = []
	GC += [0]*(steps+1)
	for i in range(steps):
		binstart = i*(int(resolution/1000))
		binend = (i+1)*(int(resolution/1000))
		averageGC = np.average(GCper[binstart:binend])
		GC[i] = averageGC
	lastbinStart = steps*(int(resolution/1000))
	lastbinEnd = len(GCper)
	GC[-1] = np.average(GCper[lastbinStart:lastbinEnd])
	#print GCper
	GC = np.asarray(GC)

	return GC

def getmappability(mappablity_track, matrixchrom, matrixstart, matrixend, resolution):
	tb = tabix.open(mappablity_track)
	records = tb.query(matrixchrom, int(matrixstart), int(matrixend))
	Allstart = int(matrixstart)
	Allend = int(matrixend)
	values = []
	binvalues = []
	values += [0]*(Allstart%resolution)
	for r in records:
		chrom,start,end,value = r
		length = int(end) - int(start)
		if int(start) == Allstart:
			temp = [float(value)]*length
		else:
			temp = [0]*(int(start)-Allstart) + [float(value)]*length
		values += temp
		Allstart = int(end)
	if Allstart == Allend:
		pass
	else:
		values += [0]*(Allend-Allstart)
	values += [0]*(resolution-Allend%resolution)
	tempcount = 0
	for i,v in enumerate(values):
		if i%resolution == (resolution-1):
			binvalues.append(float(tempcount)/float(resolution))
			tempcount = 0
		else:
			tempcount += v
	return binvalues

def getRestrictionSitesInfo(restrictionSites):
	inf = open(restrictionSites)
	sitesInfo = {}
	for line in inf:
		line = line.strip().split()
		chrom = line[0]
		sites = line[1:]
		sitesInfo[chrom] = sites
	return sitesInfo

def getFragmentsNumber(sitesInfo, matrixchrom, resolution,chromlength):
	sites = sitesInfo[matrixchrom]
	bincounts = [0.0] * (int(int(chromlength)/int(resolution)) + 1)
	for site in sites:
		idx = int(int(site) / int(resolution))
		if idx >= len(bincounts):
			pass
		else:
			bincounts[idx] += 1.0
	return bincounts

def Regression(rowsums, mappability,gc,fragmentCounts,outputfilename, filesName):
	matrix = rowsums.values
	label = filesName[0]
	label = label.lstrip('#').strip()
	outRegressionfile = outputfilename + '%s.txt'%label

	#print outRegressionfile
	rowsum = np.asarray(matrix[:,0])
	nrowdim = len(rowsum)
	GC = np.asarray(gc[:nrowdim])
	mapscore = np.asarray(mappability[:nrowdim])
	fragmentscore = np.asarray(fragmentCounts[:nrowdim])
	#print len(rowsum),len(GC),len(mapscore),len(fragmentscore)

	d = {'rowsum':rowsum, 'gcper':GC, 'mapscore': mapscore,'cutSitesNumber':fragmentscore}
	mydata = pd.DataFrame(d,columns=['cutSitesNumber','gcper','mapscore','rowsum'])
	mydata.to_csv(outRegressionfile,sep="\t",header=False)

	return outRegressionfile

def getnonzeros(outputfilename,nonzerooutputfilename):
	inf = open(outputfilename)
	outf = open(nonzerooutputfilename,'w')
	for line in inf:
		if line.startswith(' '):
			pass
		else:
			newline = line.strip().split('\t')
			if len(newline) != 5:
				pass
				#print line
			else:
				id,cn,gc,ms,rs = newline
				if float(cn) == float(gc) == float(ms) == 0:
					pass
				else:
					outf.write('\t'.join(newline) + '\n')
	outf.close()

def mergeAllchroms(regressionFileInfo,chroms,outfile,headerfile):
	command = "cat %s "%headerfile
	for chrom in chroms:
		if chrom in regressionFileInfo:
			datafile = regressionFileInfo[chrom]
			command += datafile
			command += ' '
		else:
			pass
	command += '> %s'%outfile
	print(command)
	run_cmd(command)

def prepareData(name,outdir,referencedir,chromlf,rowSumFilesInfo,binsize,hg19_1k_GCPercent,mappablity_track,restrictionSites):
	regressionFileInfo = {}
	chroms,chromInfo = get_chromInfo(chromlf)
	newchroms = []
	suboutdir = os.path.join(outdir,name+'_dataForRegression')
	if os.path.isdir(suboutdir):
		outputname = os.path.join(suboutdir,"regressionInfo")
	else:
		os.mkdir(suboutdir)
		outputname = os.path.join(suboutdir,"regressionInfo")
	sitesInfo = getRestrictionSitesInfo(restrictionSites)
	for chrom in chromInfo:
		if chrom not in rowSumFilesInfo or chrom == 'chrY' or chrom == 'chrM':
			pass
		else:
			#print chrom
			chromLength = chromInfo[chrom]
			file = rowSumFilesInfo[chrom]
			rowsums = pd.read_csv(file, header = 0,sep="\t",index_col=0)
			filesName = list(rowsums)
			#print filesName
			resolution = int(binsize)*1000
			mapscore = getmappability(mappablity_track, chrom, 1, chromLength, resolution)
			fragmentCounts = getFragmentsNumber(sitesInfo, chrom, resolution, chromLength)
			gc = getGCpercent(hg19_1k_GCPercent,chrom,resolution)
			outputfilename = outputname + '_%s_'%chrom
			outRegressionfile = Regression(rowsums, mapscore, gc, fragmentCounts, outputfilename, filesName)
			nonzerooutputfilename = outRegressionfile.replace('.txt', '_nonzero.txt')
			getnonzeros(outRegressionfile, nonzerooutputfilename)
			regressionFileInfo[chrom] = nonzerooutputfilename
			newchroms.append(chrom)
	outfile = os.path.join(outputname + '_nonZeros_allchroms.txt')
	headerfile = os.path.join(referencedir, 'regressionDataHeader.txt')
	mergeAllchroms(regressionFileInfo,chroms,outfile,headerfile)
	regressionData = outfile
	return newchroms,regressionData,regressionFileInfo



import os,sys
import numpy as np
from HiNT.corelib import *

def makebinFiles(residualf,regressionf,outbinfile,binsize):
	outf = open(outbinfile,'w')
	title = ['start','end','obs','expected','var']
	outf.write('\t'.join(title)+'\n')
	inbinf = open(regressionf)
	allbins = []
	for line in inbinf:
		line = line.strip().split('\t')
		if len(line) != 5:
			pass
		else:
			bins = int(line[0])
			allbins.append(bins)
	matrix = np.loadtxt(residualf,delimiter='\t')
	mint = np.min(matrix)
	normtreat = matrix - mint
	if np.shape(normtreat)[0] != len(allbins):
		print(np.shape(normtreat)[0],len(allbins))
		print("size of the residuals and bins are not match!")
		sys.exit()
	else:
		for i,rowsum in enumerate(normtreat):
			binID = allbins[i]
			start = binID*binsize
			end = (binID + 1)*binsize - 1
			obs = rowsum
			expected = 0 - mint
			var = obs - expected
			row = [str(start),str(end),str(obs),str(expected),str(var)]
			outf.write('\t'.join(row) + '\n')
	outf.close()

def BICseqPrepare(name,outdir,resolution,resiudalChromFilesInfo,chroms):
	subDir = os.path.join(outdir,"segmentation")
	if not os.path.isdir(subDir):
		os.mkdir(subDir)
	outputconfigfile = os.path.join(subDir,name + '_GAMregression_allchroms_%skb.cfg'%str(resolution))
	outf2 = open(outputconfigfile,'w')
	title2 = ['chrom','sample']
	outf2.write('\t'.join(title2) + '\n')
	outbinDir = os.path.join(subDir,'b%s'%str(resolution*1000))
	if not os.path.isdir(outbinDir):
		os.mkdir(outbinDir)
	for chrom in chroms:
		formatedchrom = chrom.lstrip('chr')
		if len(formatedchrom) == 1 and formatedchrom.isdigit():
			formatedchrom = '0' + formatedchrom
		if chrom not in resiudalChromFilesInfo:
			pass
		else:
			regressionf,residualf = resiudalChromFilesInfo[chrom]
			outbinfile = os.path.join(outbinDir,"%s_GAM_%skb.chrm_%s.b%s.bin"%(name,str(resolution),formatedchrom,str(resolution*1000)))
			makebinFiles(residualf,regressionf,outbinfile,resolution*1000)
			outf2.write(chrom+ '\t' + outbinfile + '\n')
	return subDir,outbinDir,outputconfigfile

def run_BICseq(name,outdir,resolution,resiudalChromFilesInfo,chroms,BICseqpath,CNVplotPath,genome):
	#print resiudalChromFilesInfo
	subDir,outbinDir,outputconfigfile = BICseqPrepare(name,outdir,resolution,resiudalChromFilesInfo,chroms)
	bicseqOut = os.path.join(subDir,name + '_CNV_segments.txt')
	tmp = os.path.join(subDir,'tmp')
	if not os.path.isdir(tmp):
		os.mkdir(tmp)
	outfig = os.path.join(subDir,name + '_%skb_CNV_segments.png'%str(resolution))
	command = 'perl %s %s %s --tmp %s --title %s_%skb_CNV --bootstrap --fig %s --lambda 3'%(BICseqpath,outputconfigfile,bicseqOut,tmp,name,str(resolution),outfig)
	print(command)
	run_cmd(command)
	#command2 = 'Rscript %s --lamda 3 --title %s_%skb_GAMPoissonResiduals -i %s -o %s --bin_size %s --chrm'%(CNVplotPath,name,str(resolution),outbinDir,bicseqOut,str(int(resolution)*1000))
	if genome.startswith('hg'):
		mouse = "0"
	if genome.startswith('mm'):
		mouse = "1"
	command2 = 'Rscript %s %s %s %s 3 %s %s_%skb_GAMPoissonResiduals --chrm'%(CNVplotPath, outbinDir, bicseqOut, str(int(resolution)*1000), mouse, name,str(resolution))
	print(command2)
	run_cmd(command2)
	outfig2 = os.path.join(subDir,name + '_CNV_segments.l2r.pdf')
	return bicseqOut,outfig,outfig2



#!/home/sw229/miniconda2/bin/env python

#version: 2.0
#author: Su Wang
#Contact: wangsu0623@gmail.com

import os,sys,io
from distutils.core import setup, Extension
from setuptools import find_packages
from os import path

if sys.version < "3.5.0":
    print("Please use a Python with higher version than 3.5.0")
    sys.stderr.write("CRITICAL: Python version must be 3.5 or higher!\n")
    sys.exit(1)
    exit(1)

this_directory = path.abspath(path.dirname(__file__))
with io.open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

#with io.open(path.join(this_directory, 'requirements.txt')) as f:
#    requires = f.read().splitlines()
#requires = [req.strip() for req in requires]

requires = [
'h5py>=2.8.0',
'numpy>=1.16.1',
'cooler>=0.8.2',
'pypairix>=0.3.0',
'pairtools>=0.2.2',
'scipy>=1.0.1',
'pandas>=0.23.0',
'scikit-learn>=0.19.1',
'multiprocess>=0.70.5',
'argparse>=1.1',
'pytabix'
]

def main():
    #compilemis()
    if not float(sys.version[:3])>=3.5:
        sys.stderr.write("CRITICAL: Python version must be greater than or equal to 3.5!\n")
        sys.exit(1)
    setup(name="HiNT-Package",
          version="2.0.5",
          description="HiNT -- HiC for copy number vairations and translocations detection ",
          long_description=long_description,
          long_description_content_type='text/markdown',
          author='Su Wang',
          author_email='wangsu0623@gmail.com',
          package_dir={'HiNT' : 'HiNT'},
          install_requires = requires,
          setup_requires = requires,
          packages=['HiNT'],
          package_data={'HiNT':['externalScripts/*']},
          include_package_data=True,
          scripts=['bin/hint'],
	  url="https://github.com/suwangbio/HiNT_py3",
          classifiers=[
            'Programming Language :: Python :: 3',
            'Environment :: Console',
            'Environment :: Web Environment',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Operating System :: MacOS :: MacOS X',
            'Operating System :: POSIX',
            'Topic :: Software Development',
            ],
          )


if __name__ == '__main__':
    main()



"""diff

Prints out where two hic files are different

First attempt; resolutions, chromosomes, norms hard-coded 
(these should instead be taken from respective hic files)
Only checks intra-chromosomal matrices
This is very slow particularly for FRAG since it looks at 
every single matrix entry
"""
from . import straw
import math
import argparse

bpresolutions=(2500000,1000000,500000,100000,50000,25000,10000,5000)
fragresolutions=(500,200,100,50,20,5,2,1) 
chrs=('1','2','3','4','5','6','7','8','9','10','11','12','13','14','15','16','17','18','19','20','21','22','X')
norms=('NONE', 'KR', 'VC')

def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    '''
    Python 2 implementation of Python 3.5 math.isclose()
    https://hg.python.org/cpython/file/tip/Modules/mathmodule.c#l1993
    '''
    # sanity check on the inputs
    if rel_tol < 0 or abs_tol < 0:
        raise ValueError("tolerances must be non-negative")

    # short circuit exact equality -- needed to catch two infinities of
    # the same sign. And perhaps speeds things up a bit sometimes.
    if a == b:
        return True

    # This catches the case of two infinities of opposite sign, or
    # one infinity and one finite number. Two infinities of opposite
    # sign would otherwise have an infinite relative tolerance.
    # Two infinities of the same sign are caught by the equality check
    # above.
    if math.isinf(a) or math.isinf(b):
        return False

    # equality check above would return false for nan, but we want
    # to return true
    if math.isnan(a) and math.isnan(b):
        return True

    # now do the regular computation
    # this is essentially the "weak" test from the Boost library
    diff = math.fabs(b - a)
    result = (((diff <= math.fabs(rel_tol * b)) or
               (diff <= math.fabs(rel_tol * a))) or
              (diff <= abs_tol))
    return result


parser = argparse.ArgumentParser(description='Print differences between hic files, or nothing if they are the same')
parser.add_argument('filenames', metavar='file', nargs=2, help='files to diff')
parser.add_argument('--frag_only', dest='frag_only', default=False, action='store_true', help='only examine BP resolutions')
parser.add_argument('--bp_only', dest='bp_only', default=False, action='store_true', help='only examine frag resolutions')
args = parser.parse_args()
file1=args.filenames[0]
file2=args.filenames[1]

if (not args.frag_only):
    for res in bpresolutions:
        for chr in chrs:
            for norm in norms:
                result=straw.straw(norm, file1, chr, chr, "BP", res)
                result2=straw.straw(norm, file2, chr, chr, "BP", res)
                for i in range(len(result[0])):
                    if (not isclose(result[0][i],result2[0][i]) or not isclose(result[1][i],result2[1][i]) or not isclose(result[2][i], result2[2][i])):
                        print(("Difference at {0} {1} BP {2}".format(norm, chr, res)))
                        print(("{0}\t{1}\t{2}\t{3}\t{4}\t{5}".format(result[0][i], result[1][i], result[2][i], result2[0][i], result2[1][i], result2[2][i])))

if (not args.bp_only):
    for res in fragresolutions:
        for chr in chrs:
            for norm in norms:
                result=straw.straw(norm, file1, chr, chr, "FRAG", res)
                result2=straw.straw(norm, file2, chr, chr, "FRAG", res)
                for i in range(len(result[0])):
                    if (not isclose(result[0][i],result2[0][i]) or not isclose(result[1][i],result2[1][i]) or not isclose(result[2][i], result2[2][i])):
                        print(("Difference at {0} {1} FRAG {2}".format(norm, chr, res)))
                        print(("{0}\t{1}\t{2}\t{3}\t{4}\t{5}".format(result[0][i], result[1][i], result[2][i], result2[0][i], result2[1][i], result2[2][i])))
